from __future__ import annotations

import random

import importlib_resources
from chik_rs import AugSchemeMPL, ClassgroupElement, Coin, G1Element, G2Element, VDFInfo, VDFProof

from chik.consensus.coinbase import create_farmer_coin, create_pool_coin
from chik.consensus.default_constants import DEFAULT_CONSTANTS
from chik.types.blockchain_format.foliage import Foliage, FoliageBlockData, FoliageTransactionBlock, TransactionsInfo
from chik.types.blockchain_format.pool_target import PoolTarget
from chik.types.blockchain_format.proof_of_space import ProofOfSpace
from chik.types.blockchain_format.reward_chain_block import RewardChainBlock
from chik.types.blockchain_format.serialized_program import SerializedProgram
from chik.types.blockchain_format.sized_bytes import bytes32, bytes100
from chik.types.full_block import FullBlock
from chik.util.ints import uint8, uint32, uint64, uint128

# farmer puzzle hash
ph = bytes32(b"a" * 32)

klvm_generator_bin_path = importlib_resources.files(__name__.rpartition(".")[0]).joinpath("klvm_generator.bin")
klvm_generator = klvm_generator_bin_path.read_bytes()


def rewards(height: uint32) -> tuple[Coin, Coin]:
    farmer_coin = create_farmer_coin(height, ph, uint64(250000000), DEFAULT_CONSTANTS.GENESIS_CHALLENGE)
    pool_coin = create_pool_coin(height, ph, uint64(1750000000), DEFAULT_CONSTANTS.GENESIS_CHALLENGE)
    return farmer_coin, pool_coin


def rand_bytes(num: int) -> bytes:
    ret = bytearray(num)
    for i in range(num):
        ret[i] = random.getrandbits(8)
    return bytes(ret)


def rand_hash() -> bytes32:
    return bytes32(rand_bytes(32))


def rand_g1() -> G1Element:
    sk = AugSchemeMPL.key_gen(rand_bytes(96))
    return sk.get_g1()


def rand_g2() -> G2Element:
    sk = AugSchemeMPL.key_gen(rand_bytes(96))
    return AugSchemeMPL.sign(sk, b"foobar")


def rand_class_group_element() -> ClassgroupElement:
    return ClassgroupElement(bytes100(rand_bytes(100)))


def rand_vdf() -> VDFInfo:
    return VDFInfo(rand_hash(), uint64(random.randint(100000, 1000000000)), rand_class_group_element())


def rand_vdf_proof() -> VDFProof:
    return VDFProof(
        uint8(1),  # witness_type
        rand_hash(),  # witness
        bool(random.randint(0, 1)),  # normalized_to_identity
    )


def rand_full_block() -> FullBlock:
    proof_of_space = ProofOfSpace(
        rand_hash(),
        rand_g1(),
        None,
        rand_g1(),
        uint8(0),
        rand_bytes(8 * 32),
    )

    reward_chain_block = RewardChainBlock(
        uint128(1),
        uint32(2),
        uint128(3),
        uint8(4),
        rand_hash(),
        proof_of_space,
        None,
        rand_g2(),
        rand_vdf(),
        None,
        rand_g2(),
        rand_vdf(),
        rand_vdf(),
        True,
    )

    pool_target = PoolTarget(
        rand_hash(),
        uint32(0),
    )

    foliage_block_data = FoliageBlockData(
        rand_hash(),
        pool_target,
        rand_g2(),
        rand_hash(),
        rand_hash(),
    )

    foliage = Foliage(
        rand_hash(),
        rand_hash(),
        foliage_block_data,
        rand_g2(),
        rand_hash(),
        rand_g2(),
    )

    foliage_transaction_block = FoliageTransactionBlock(
        rand_hash(),
        uint64(0),
        rand_hash(),
        rand_hash(),
        rand_hash(),
        rand_hash(),
    )

    farmer_coin, pool_coin = rewards(uint32(0))

    transactions_info = TransactionsInfo(
        rand_hash(),
        rand_hash(),
        rand_g2(),
        uint64(0),
        uint64(1),
        [farmer_coin, pool_coin],
    )

    full_block = FullBlock(
        [],
        reward_chain_block,
        rand_vdf_proof(),
        rand_vdf_proof(),
        rand_vdf_proof(),
        rand_vdf_proof(),
        rand_vdf_proof(),
        foliage,
        foliage_transaction_block,
        transactions_info,
        SerializedProgram.from_bytes(klvm_generator),
        [],
    )

    return full_block



from __future__ import annotations

from typing import Any, Optional

import pytest
from chik_rs import AugSchemeMPL
from klvm.casts import int_to_bytes

from chik._tests.util.spend_sim import SimClient, SpendSim, sim_and_client
from chik.types.blockchain_format.coin import Coin
from chik.types.blockchain_format.program import INFINITE_COST, Program
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.coin_spend import make_spend
from chik.types.condition_opcodes import ConditionOpcode
from chik.types.mempool_inclusion_status import MempoolInclusionStatus
from chik.util.condition_tools import conditions_dict_for_solution
from chik.util.errors import Err
from chik.util.hash import std_hash
from chik.util.ints import uint32, uint64
from chik.wallet.cat_wallet.cat_utils import CAT_MOD
from chik.wallet.dao_wallet.dao_info import DAORules
from chik.wallet.dao_wallet.dao_utils import curry_singleton, get_p2_singleton_puzhash, get_treasury_puzzle
from chik.wallet.puzzles.load_klvm import load_klvm
from chik.wallet.singleton import create_singleton_puzzle_hash
from chik.wallet.wallet_spend_bundle import WalletSpendBundle

CAT_MOD_HASH: bytes32 = CAT_MOD.get_tree_hash()
SINGLETON_MOD: Program = load_klvm("singleton_top_layer_v1_1.clsp")
SINGLETON_MOD_HASH: bytes32 = SINGLETON_MOD.get_tree_hash()
SINGLETON_LAUNCHER: Program = load_klvm("singleton_launcher.clsp")
SINGLETON_LAUNCHER_HASH: bytes32 = SINGLETON_LAUNCHER.get_tree_hash()
DAO_LOCKUP_MOD: Program = load_klvm("dao_lockup.clsp")
DAO_LOCKUP_MOD_HASH: bytes32 = DAO_LOCKUP_MOD.get_tree_hash()
DAO_PROPOSAL_TIMER_MOD: Program = load_klvm("dao_proposal_timer.clsp")
DAO_PROPOSAL_TIMER_MOD_HASH: bytes32 = DAO_PROPOSAL_TIMER_MOD.get_tree_hash()
DAO_PROPOSAL_MOD: Program = load_klvm("dao_proposal.clsp")
DAO_PROPOSAL_MOD_HASH: bytes32 = DAO_PROPOSAL_MOD.get_tree_hash()
DAO_PROPOSAL_VALIDATOR_MOD: Program = load_klvm("dao_proposal_validator.clsp")
DAO_PROPOSAL_VALIDATOR_MOD_HASH: bytes32 = DAO_PROPOSAL_VALIDATOR_MOD.get_tree_hash()
DAO_TREASURY_MOD: Program = load_klvm("dao_treasury.clsp")
DAO_TREASURY_MOD_HASH: bytes32 = DAO_TREASURY_MOD.get_tree_hash()
SPEND_P2_SINGLETON_MOD: Program = load_klvm("dao_spend_p2_singleton_v2.clsp")
SPEND_P2_SINGLETON_MOD_HASH: bytes32 = SPEND_P2_SINGLETON_MOD.get_tree_hash()
DAO_FINISHED_STATE: Program = load_klvm("dao_finished_state.clsp")
DAO_FINISHED_STATE_HASH: bytes32 = DAO_FINISHED_STATE.get_tree_hash()
DAO_CAT_TAIL: Program = load_klvm(
    "genesis_by_coin_id_or_singleton.clsp", package_or_requirement="chik.wallet.cat_wallet.puzzles"
)
DAO_CAT_TAIL_HASH: bytes32 = DAO_CAT_TAIL.get_tree_hash()
P2_SINGLETON_MOD: Program = load_klvm("p2_singleton_via_delegated_puzzle.clsp")
P2_SINGLETON_MOD_HASH: bytes32 = P2_SINGLETON_MOD.get_tree_hash()
P2_SINGLETON_AGGREGATOR_MOD: Program = load_klvm("p2_singleton_aggregator.clsp")
P2_SINGLETON_AGGREGATOR_MOD_HASH: bytes32 = P2_SINGLETON_AGGREGATOR_MOD.get_tree_hash()
DAO_UPDATE_MOD: Program = load_klvm("dao_update_proposal.clsp")
DAO_UPDATE_MOD_HASH: bytes32 = DAO_UPDATE_MOD.get_tree_hash()


def test_finished_state() -> None:
    """
    Once a proposal has closed, it becomes a 'beacon' singleton which announces
    its proposal ID. This is referred to as the finished state and is used to
    confirm that a proposal has closed in order to release voting CATs from
    the lockup puzzle.
    """
    proposal_id = Program.to("proposal_id").get_tree_hash()
    singleton_struct: Program = Program.to(
        (SINGLETON_MOD.get_tree_hash(), (proposal_id, SINGLETON_LAUNCHER.get_tree_hash()))
    )
    finished_inner_puz = DAO_FINISHED_STATE.curry(singleton_struct, DAO_FINISHED_STATE_HASH)
    finished_full_puz = SINGLETON_MOD.curry(singleton_struct, finished_inner_puz)
    inner_sol = Program.to([1])

    conds = finished_inner_puz.run(inner_sol).as_python()
    assert conds[0][1] == finished_full_puz.get_tree_hash()
    assert conds[2][1] == finished_inner_puz.get_tree_hash()

    lineage = Program.to([proposal_id, finished_inner_puz.get_tree_hash(), 1])
    full_sol = Program.to([lineage, 1, inner_sol])

    conds = conditions_dict_for_solution(finished_full_puz, full_sol, INFINITE_COST)
    assert conds[ConditionOpcode.ASSERT_MY_PUZZLEHASH][0].vars[0] == finished_full_puz.get_tree_hash()
    assert conds[ConditionOpcode.CREATE_COIN][0].vars[0] == finished_full_puz.get_tree_hash()


def test_proposal() -> None:
    """
    This test covers the three paths for closing a proposal:
    - Close a passed proposal
    - Close a failed proposal
    - Self-destruct a broken proposal
    """
    proposal_pass_percentage: uint64 = uint64(5100)
    CAT_TAIL_HASH = Program.to("tail").get_tree_hash()
    treasury_id = Program.to("treasury").get_tree_hash()
    singleton_id = Program.to("singleton_id").get_tree_hash()
    singleton_struct: Program = Program.to(
        (SINGLETON_MOD.get_tree_hash(), (singleton_id, SINGLETON_LAUNCHER.get_tree_hash()))
    )
    self_destruct_time = 1000  # number of blocks
    oracle_spend_delay = 10
    active_votes_list = [0xFADEDDAB]  # are the ids of previously voted on proposals?
    acs: Program = Program.to(1)
    acs_ph: bytes32 = acs.get_tree_hash()

    dao_lockup_self = DAO_LOCKUP_MOD.curry(
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        DAO_FINISHED_STATE_HASH,
        CAT_MOD_HASH,
        CAT_TAIL_HASH,
    )

    proposal_curry_one = DAO_PROPOSAL_MOD.curry(
        DAO_PROPOSAL_TIMER_MOD_HASH,
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        CAT_MOD_HASH,
        DAO_FINISHED_STATE_HASH,
        DAO_TREASURY_MOD_HASH,
        dao_lockup_self.get_tree_hash(),
        CAT_TAIL_HASH,
        treasury_id,
    )

    # make a lockup puz for the dao cat
    lockup_puz = dao_lockup_self.curry(
        dao_lockup_self.get_tree_hash(),
        active_votes_list,
        acs,  # innerpuz
    )

    dao_cat_puz: Program = CAT_MOD.curry(CAT_MOD_HASH, CAT_TAIL_HASH, lockup_puz)
    dao_cat_puzhash: bytes32 = dao_cat_puz.get_tree_hash()

    # Test Voting
    current_yes_votes = 20
    current_total_votes = 100
    full_proposal: Program = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        singleton_id,
        acs_ph,
        current_yes_votes,
        current_total_votes,
    )

    vote_amount = 10
    vote_type = 1  # yes vote
    vote_coin_id = Program.to("vote_coin").get_tree_hash()
    solution: Program = Program.to(
        [
            [vote_amount],  # vote amounts
            vote_type,  # vote type (yes)
            [vote_coin_id],  # vote coin ids
            [active_votes_list],  # previous votes
            [acs_ph],  # lockup inner puz hash
            0,  # inner puz reveal
            0,  # soft close len
            self_destruct_time,
            oracle_spend_delay,
            0,
            1,
        ]
    )

    # Run the proposal and check its conditions
    conditions = conditions_dict_for_solution(full_proposal, solution, INFINITE_COST)

    # Puzzle Announcement of vote_coin_ids
    assert bytes32(conditions[ConditionOpcode.CREATE_PUZZLE_ANNOUNCEMENT][0].vars[0]) == vote_coin_id

    # Assert puzzle announcement from dao_cat of proposal_id and all vote details
    apa_msg = Program.to([singleton_id, vote_amount, vote_type, vote_coin_id]).get_tree_hash()
    assert conditions[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT][0].vars[0] == std_hash(dao_cat_puzhash + apa_msg)

    # Check that the proposal recreates itself with updated vote amounts
    next_proposal: Program = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        singleton_id,
        acs_ph,
        current_yes_votes + vote_amount,
        current_total_votes + vote_amount,
    )
    assert bytes32(conditions[ConditionOpcode.CREATE_COIN][0].vars[0]) == next_proposal.get_tree_hash()
    assert conditions[ConditionOpcode.CREATE_COIN][0].vars[1] == int_to_bytes(1)

    # Try to vote using multiple coin ids
    vote_coin_id_1 = Program.to("vote_coin_1").get_tree_hash()
    vote_coin_id_2 = Program.to("vote_coin_2").get_tree_hash()
    repeat_solution_1: Program = Program.to(
        [
            [vote_amount, 20],  # vote amounts
            vote_type,  # vote type (yes)
            [vote_coin_id_1, vote_coin_id_2],  # vote coin ids
            [active_votes_list, 0],  # previous votes
            [acs_ph, acs_ph],  # lockup inner puz hash
            0,  # inner puz reveal
            0,  # soft close len
            self_destruct_time,
            oracle_spend_delay,
            0,
            1,
        ]
    )

    conds_repeated = conditions_dict_for_solution(full_proposal, repeat_solution_1, INFINITE_COST)
    assert len(conds_repeated) == 5

    # Try to vote using repeated coin ids
    repeat_solution_2: Program = Program.to(
        [
            [vote_amount, vote_amount, 20],  # vote amounts
            vote_type,  # vote type (yes)
            [vote_coin_id_1, vote_coin_id_1, vote_coin_id_2],  # vote coin ids
            [active_votes_list],  # previous votes
            [acs_ph],  # lockup inner puz hash
            0,  # inner puz reveal
            0,  # soft close len
            self_destruct_time,
            oracle_spend_delay,
            0,
            1,
        ]
    )

    with pytest.raises(ValueError) as e_info:
        conditions_dict_for_solution(full_proposal, repeat_solution_2, INFINITE_COST)
    assert "klvm raise" in e_info.value.args[0]

    # Test Launch
    current_yes_votes = 0
    current_total_votes = 0
    launch_proposal: Program = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        singleton_id,
        acs_ph,
        current_yes_votes,
        current_total_votes,
    )
    vote_amount = 10
    vote_type = 1  # yes vote
    vote_coin_id = Program.to("vote_coin").get_tree_hash()
    solution = Program.to(
        [
            [vote_amount],  # vote amounts
            vote_type,  # vote type (yes)
            [vote_coin_id],  # vote coin ids
            # TODO: Check whether previous votes should be 0 in the first spend since
            # proposal looks at (f previous_votes) during loop_over_vote_coins
            [0],  # previous votes
            [acs_ph],  # lockup inner puz hash
            acs,  # inner puz reveal
            0,  # soft close len
            self_destruct_time,
            oracle_spend_delay,
            0,
            1,
        ]
    )
    # Run the proposal and check its conditions
    conditions = conditions_dict_for_solution(launch_proposal, solution, INFINITE_COST)
    # check that the timer is created
    timer_puz = DAO_PROPOSAL_TIMER_MOD.curry(
        proposal_curry_one.get_tree_hash(),
        singleton_struct,
    )
    timer_puzhash = timer_puz.get_tree_hash()
    assert conditions[ConditionOpcode.CREATE_COIN][1].vars[0] == timer_puzhash

    # Test exits

    # Test attempt to close a passing proposal
    current_yes_votes = 200
    current_total_votes = 350
    full_proposal = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        singleton_id,
        acs_ph,
        current_yes_votes,
        current_total_votes,
    )
    attendance_required = 200
    proposal_timelock = 20
    soft_close_length = 5
    solution = Program.to(
        [
            Program.to("validator_hash").get_tree_hash(),
            0,
            # Program.to("receiver_hash").get_tree_hash(), # not needed anymore?
            proposal_timelock,
            proposal_pass_percentage,
            attendance_required,
            0,
            soft_close_length,
            self_destruct_time,
            oracle_spend_delay,
            0,
            1,
        ]
    )

    conds = conditions_dict_for_solution(full_proposal, solution, INFINITE_COST)

    # make a matching treasury puzzle for the APA
    treasury_inner: Program = DAO_TREASURY_MOD.curry(
        DAO_TREASURY_MOD_HASH,
        Program.to("validator_hash"),
        proposal_timelock,
        soft_close_length,
        attendance_required,
        proposal_pass_percentage,
        self_destruct_time,
        oracle_spend_delay,
    )
    treasury: Program = SINGLETON_MOD.curry(
        Program.to((SINGLETON_MOD_HASH, (treasury_id, SINGLETON_LAUNCHER_HASH))),
        treasury_inner,
    )
    treasury_puzhash = treasury.get_tree_hash()
    apa_msg_bytes = singleton_id

    timer_apa = std_hash(timer_puzhash + singleton_id)
    assert conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT][0].vars[0] == timer_apa
    assert conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT][1].vars[0] == std_hash(treasury_puzhash + apa_msg_bytes)

    # close a failed proposal
    full_proposal = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        singleton_id,
        acs_ph,
        20,  # failing number of yes votes
        current_total_votes,
    )
    solution = Program.to(
        [
            Program.to("validator_hash").get_tree_hash(),
            0,
            # Program.to("receiver_hash").get_tree_hash(), # not needed anymore?
            proposal_timelock,
            proposal_pass_percentage,
            attendance_required,
            0,
            soft_close_length,
            self_destruct_time,
            oracle_spend_delay,
            0,
            1,
        ]
    )
    conds = conditions_dict_for_solution(full_proposal, solution, INFINITE_COST)
    apa_msg_int = int_to_bytes(0)
    assert conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT][1].vars[0] == std_hash(treasury_puzhash + apa_msg_int)
    assert conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT][0].vars[0] == timer_apa

    finished_puz = DAO_FINISHED_STATE.curry(singleton_struct, DAO_FINISHED_STATE_HASH)
    assert conds[ConditionOpcode.CREATE_COIN][0].vars[0] == finished_puz.get_tree_hash()

    # self destruct a proposal
    attendance_required = 200
    solution = Program.to(
        [
            Program.to("validator_hash").get_tree_hash(),
            0,
            # Program.to("receiver_hash").get_tree_hash(), # not needed anymore?
            proposal_timelock,
            proposal_pass_percentage,
            attendance_required,
            0,
            soft_close_length,
            self_destruct_time,
            oracle_spend_delay,
            1,
            1,
        ]
    )
    conds = conditions_dict_for_solution(full_proposal, solution, INFINITE_COST)
    assert conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT][0].vars[0] == std_hash(treasury_puzhash + apa_msg_int)
    assert conds[ConditionOpcode.CREATE_COIN][0].vars[0] == finished_puz.get_tree_hash()


def test_proposal_timer() -> None:
    """
    The timer puzzle is created at the same time as a proposal, and enforces a relative time condition on proposals
    The closing time is passed in via the timer solution and confirmed via announcement from the proposal.
    It creates/asserts announcements to pair it with the finishing spend of a proposal.
    The timer puzzle only has one spend path so there is only one test case for this puzzle.
    """
    CAT_TAIL_HASH = Program.to("tail").get_tree_hash()
    treasury_id = Program.to("treasury").get_tree_hash()
    singleton_id = Program.to("singleton_id").get_tree_hash()
    singleton_struct: Program = Program.to(
        (SINGLETON_MOD.get_tree_hash(), (singleton_id, SINGLETON_LAUNCHER.get_tree_hash()))
    )
    dao_lockup_self = DAO_LOCKUP_MOD.curry(
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        DAO_FINISHED_STATE_HASH,
        CAT_MOD_HASH,
        CAT_TAIL_HASH,
    )

    proposal_curry_one = DAO_PROPOSAL_MOD.curry(
        DAO_PROPOSAL_TIMER_MOD_HASH,
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        CAT_MOD_HASH,
        DAO_FINISHED_STATE_HASH,
        DAO_TREASURY_MOD_HASH,
        dao_lockup_self.get_tree_hash(),
        CAT_TAIL_HASH,
        treasury_id,
    )

    proposal_timer_full: Program = DAO_PROPOSAL_TIMER_MOD.curry(
        proposal_curry_one.get_tree_hash(),
        singleton_struct,
    )

    timelock = int_to_bytes(101)
    parent_parent_id = Program.to("parent_parent").get_tree_hash()
    parent_amount = 2000
    solution: Program = Program.to(
        [
            140,  # yes votes
            180,  # total votes
            Program.to(1).get_tree_hash(),  # proposal innerpuz
            timelock,
            parent_parent_id,
            parent_amount,
        ]
    )
    # run the timer puzzle.
    conds = conditions_dict_for_solution(proposal_timer_full, solution, INFINITE_COST)
    assert len(conds) == 4

    # Validate the output conditions
    # Check the timelock is present
    assert conds[ConditionOpcode.ASSERT_HEIGHT_RELATIVE][0].vars[0] == timelock
    # Check the proposal id is announced by the timer puz
    assert conds[ConditionOpcode.CREATE_PUZZLE_ANNOUNCEMENT][0].vars[0] == singleton_id
    # Check the proposal puz announces the timelock
    expected_proposal_puzhash: bytes32 = create_singleton_puzzle_hash(
        proposal_curry_one.curry(
            proposal_curry_one.get_tree_hash(), singleton_id, Program.to(1).get_tree_hash(), 140, 180
        ).get_tree_hash(),
        singleton_id,
    )
    assert conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT][0].vars[0] == std_hash(
        expected_proposal_puzhash + timelock
    )
    # Check the parent is a proposal
    expected_parent_puzhash: bytes32 = create_singleton_puzzle_hash(
        proposal_curry_one.curry(
            proposal_curry_one.get_tree_hash(),
            singleton_id,
            Program.to(1).get_tree_hash(),
            0,
            0,
        ).get_tree_hash(),
        singleton_id,
    )
    parent_id = std_hash(parent_parent_id + expected_parent_puzhash + int_to_bytes(parent_amount))
    assert conds[ConditionOpcode.ASSERT_MY_PARENT_ID][0].vars[0] == parent_id


def test_validator() -> None:
    """
    The proposal validator is run by the treasury when a passing proposal is closed.
    Its main purpose is to check that the proposal's vote amounts adehere to
    the DAO rules contained in the treasury (which are passed in from the
    treasury as Truth values). It creates a puzzle announcement of the
    proposal ID, that the proposal itself asserts. It also spends the value
    held in the proposal to the excess payout puzhash.

    The test cases covered are:
    - Executing a spend proposal in which the validator executes the spend of a
      `spend_p2_singleton` coin. This is just a proposal that spends some the treasury
    - Executing an update proposal that changes the DAO rules.
    """
    # Setup the treasury
    treasury_id = Program.to("treasury_id").get_tree_hash()
    treasury_struct: Program = Program.to((SINGLETON_MOD_HASH, (treasury_id, SINGLETON_LAUNCHER_HASH)))

    # Setup the proposal
    proposal_id = Program.to("proposal_id").get_tree_hash()
    proposal_struct: Program = Program.to((SINGLETON_MOD.get_tree_hash(), (proposal_id, SINGLETON_LAUNCHER_HASH)))
    CAT_TAIL_HASH = Program.to("tail").get_tree_hash()
    acs: Program = Program.to(1)
    acs_ph: bytes32 = acs.get_tree_hash()

    p2_singleton = P2_SINGLETON_MOD.curry(treasury_struct, P2_SINGLETON_AGGREGATOR_MOD)
    p2_singleton_puzhash = p2_singleton.get_tree_hash()
    parent_id = Program.to("parent").get_tree_hash()
    locked_amount = 100000
    spend_amount = 1100
    conditions = [[51, 0xDABBAD00, 1000], [51, 0xCAFEF00D, 100]]

    # Setup the validator
    minimum_amt = 1
    excess_puzhash = bytes32(b"1" * 32)
    dao_lockup_self = DAO_LOCKUP_MOD.curry(
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        DAO_FINISHED_STATE_HASH,
        CAT_MOD_HASH,
        CAT_TAIL_HASH,
    )

    proposal_curry_one = DAO_PROPOSAL_MOD.curry(
        DAO_PROPOSAL_TIMER_MOD_HASH,
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        CAT_MOD_HASH,
        DAO_FINISHED_STATE_HASH,
        DAO_TREASURY_MOD_HASH,
        dao_lockup_self.get_tree_hash(),
        CAT_TAIL_HASH,
        treasury_id,
    )
    proposal_validator = DAO_PROPOSAL_VALIDATOR_MOD.curry(
        treasury_struct,
        proposal_curry_one.get_tree_hash(),
        minimum_amt,
        excess_puzhash,
    )

    # Can now create the treasury inner puz
    treasury_inner = DAO_TREASURY_MOD.curry(
        DAO_TREASURY_MOD_HASH,
        proposal_validator,
        10,  # proposal len
        5,  # soft close
        1000,  # attendance
        5100,  # pass margin
        20,  # self_destruct len
        3,  # oracle delay
    )

    # Setup the spend_p2_singleton (proposal inner puz)
    spend_p2_singleton = SPEND_P2_SINGLETON_MOD.curry(
        treasury_struct,
        CAT_MOD_HASH,
        conditions,
        [],
        p2_singleton_puzhash,  # tailhash conds
    )
    spend_p2_singleton_puzhash = spend_p2_singleton.get_tree_hash()

    parent_amt_list = [[parent_id, locked_amount]]
    cat_parent_amt_list: list[Optional[Any]] = []
    spend_p2_singleton_solution = Program.to([parent_amt_list, cat_parent_amt_list, treasury_inner.get_tree_hash()])

    output_conds = spend_p2_singleton.run(spend_p2_singleton_solution)

    proposal: Program = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        proposal_id,
        spend_p2_singleton_puzhash,
        950,
        1200,
    )
    full_proposal = SINGLETON_MOD.curry(proposal_struct, proposal)
    proposal_amt = uint64(10)
    proposal_coin_id = Coin(parent_id, full_proposal.get_tree_hash(), proposal_amt).name()
    solution = Program.to(
        [
            1000,
            5100,
            [proposal_coin_id, spend_p2_singleton_puzhash, 0],
            [proposal_id, 1200, 950, parent_id, proposal_amt],
            output_conds,
        ]
    )

    conds: Program = proposal_validator.run(solution)
    assert len(conds.as_python()) == 7 + len(conditions)

    # test update
    proposal = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        proposal_id,
        acs_ph,
        950,
        1200,
    )
    full_proposal = SINGLETON_MOD.curry(proposal_struct, proposal)
    proposal_coin_id = Coin(parent_id, full_proposal.get_tree_hash(), proposal_amt).name()
    solution = Program.to(
        [
            1000,
            5100,
            [proposal_coin_id, acs_ph, 0],
            [proposal_id, 1200, 950, parent_id, proposal_amt],
            [[51, 0xCAFEF00D, spend_amount]],
        ]
    )
    conds = proposal_validator.run(solution)
    assert len(conds.as_python()) == 3


def test_spend_p2_singleton() -> None:
    # Curried values
    singleton_id = Program.to("singleton_id").get_tree_hash()
    singleton_struct: Program = Program.to((SINGLETON_MOD_HASH, (singleton_id, SINGLETON_LAUNCHER_HASH)))
    p2_singleton_puzhash = P2_SINGLETON_MOD.curry(singleton_struct, P2_SINGLETON_AGGREGATOR_MOD).get_tree_hash()
    cat_tail_1 = Program.to("cat_tail_1").get_tree_hash()
    cat_tail_2 = Program.to("cat_tail_2").get_tree_hash()
    conditions = [[51, 0xCAFEF00D, 100], [51, 0xFEEDBEEF, 200]]
    list_of_tailhash_conds = [
        [cat_tail_1, [[51, 0x8BADF00D, 123], [51, 0xF00DF00D, 321]]],
        [cat_tail_2, [[51, 0x8BADF00D, 123], [51, 0xF00DF00D, 321]]],
    ]

    # Solution Values
    xck_parent_amt_list = [[b"x" * 32, 10], [b"y" * 32, 100]]
    cat_parent_amt_list = [
        [cat_tail_1, [["b" * 32, 100], [b"c" * 32, 400]]],
        [cat_tail_2, [[b"e" * 32, 100], [b"f" * 32, 400]]],
    ]
    # cat_parent_amt_list = []
    treasury_inner_puzhash = Program.to("treasury_inner").get_tree_hash()

    # Puzzle
    spend_p2_puz = SPEND_P2_SINGLETON_MOD.curry(
        singleton_struct, CAT_MOD_HASH, conditions, list_of_tailhash_conds, p2_singleton_puzhash
    )

    # Solution
    spend_p2_sol = Program.to([xck_parent_amt_list, cat_parent_amt_list, treasury_inner_puzhash])

    conds = spend_p2_puz.run(spend_p2_sol)
    assert conds

    # spend only cats
    conditions = []
    list_of_tailhash_conds = [
        [cat_tail_1, [[51, b"q" * 32, 123], [51, b"w" * 32, 321]]],
        [cat_tail_2, [[51, b"e" * 32, 123], [51, b"r" * 32, 321]]],
    ]
    xck_parent_amt_list = []
    cat_parent_amt_list = [
        [cat_tail_1, [[b"b" * 32, 100], [b"c" * 32, 400]]],
        [cat_tail_2, [[b"e" * 32, 100], [b"f" * 32, 400]]],
    ]
    treasury_inner_puzhash = Program.to("treasury_inner").get_tree_hash()

    # Puzzle
    spend_p2_puz = SPEND_P2_SINGLETON_MOD.curry(
        singleton_struct, CAT_MOD_HASH, conditions, list_of_tailhash_conds, p2_singleton_puzhash
    )

    # Solution
    spend_p2_sol = Program.to([xck_parent_amt_list, cat_parent_amt_list, treasury_inner_puzhash])
    conds = spend_p2_puz.run(spend_p2_sol)
    assert conds

    # test deduplicate cat_parent_amount_list
    cat_parent_amt_list = [
        [cat_tail_1, [[b"b" * 32, 100], [b"c" * 32, 400], [b"b" * 32, 100], [b"b" * 32, 100]]],
        [cat_tail_2, [[b"e" * 32, 100], [b"f" * 32, 400], [b"f" * 32, 400], [b"e" * 32, 100]]],
    ]

    spend_p2_sol = Program.to([xck_parent_amt_list, cat_parent_amt_list, treasury_inner_puzhash])
    dupe_conds = spend_p2_puz.run(spend_p2_sol)
    assert dupe_conds == conds


def test_merge_p2_singleton() -> None:
    """
    The treasury funds are held by `p2_singleton_via_delegated` puzzles.
    Because a DAO can have a large number of these coins, it's possible to
    merge them together without requiring a treasury spend.
    There are two cases tested:
    - For the merge coins that do not create the single output coin, and
    - For the coin that does create the output.
    """
    # Setup a singleton struct
    singleton_inner: Program = Program.to(1)
    singleton_id = Program.to("singleton_id").get_tree_hash()
    singleton_struct: Program = Program.to((SINGLETON_MOD_HASH, (singleton_id, SINGLETON_LAUNCHER_HASH)))

    # Setup p2_singleton_via_delegated puz
    my_id = Program.to("my_id").get_tree_hash()
    p2_singleton = P2_SINGLETON_MOD.curry(singleton_struct, P2_SINGLETON_AGGREGATOR_MOD)
    my_puzhash = p2_singleton.get_tree_hash()

    # Spend to delegated puz
    delegated_puz = Program.to(1)
    delegated_sol = Program.to([[51, 0xCAFEF00D, 300]])
    solution = Program.to([0, singleton_inner.get_tree_hash(), delegated_puz, delegated_sol, my_id])
    conds = conditions_dict_for_solution(p2_singleton, solution, INFINITE_COST)
    apa = std_hash(
        SINGLETON_MOD.curry(singleton_struct, singleton_inner).get_tree_hash()
        + Program.to([my_id, delegated_puz.get_tree_hash()]).get_tree_hash()
    )
    assert len(conds) == 4
    assert conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT][0].vars[0] == apa
    assert conds[ConditionOpcode.CREATE_COIN][0].vars[1] == int_to_bytes(300)

    # Merge Spend (not output creator)
    output_parent_id = Program.to("output_parent").get_tree_hash()
    output_coin_amount = uint64(100)
    aggregator_sol = Program.to([my_id, my_puzhash, 300, 0, [output_parent_id, output_coin_amount]])
    merge_p2_singleton_sol = Program.to([aggregator_sol, 0, 0, 0, 0])
    conds = conditions_dict_for_solution(p2_singleton, merge_p2_singleton_sol, INFINITE_COST)
    assert len(conds) == 5
    assert conds[ConditionOpcode.ASSERT_MY_PUZZLEHASH][0].vars[0] == my_puzhash
    assert conds[ConditionOpcode.CREATE_COIN_ANNOUNCEMENT][0].vars[0] == int_to_bytes(0)

    # Merge Spend (output creator)
    fake_parent_id = Program.to("fake_parent").get_tree_hash()
    merged_coin_id = Coin(fake_parent_id, my_puzhash, uint64(200)).name()
    merge_sol = Program.to([[my_id, my_puzhash, 100, [[fake_parent_id, my_puzhash, 200]], 0]])
    conds = conditions_dict_for_solution(p2_singleton, merge_sol, INFINITE_COST)
    assert len(conds) == 7
    assert conds[ConditionOpcode.ASSERT_COIN_ANNOUNCEMENT][0].vars[0] == std_hash(merged_coin_id)
    assert conds[ConditionOpcode.CREATE_COIN][0].vars[1] == int_to_bytes(300)

    # Test merge and aggregate announcements match up
    parent_ids = [
        Program.to("fake_parent_1").get_tree_hash(),
        Program.to("fake_parent_2").get_tree_hash(),
        Program.to("fake_parent_3").get_tree_hash(),
    ]
    amounts = [uint64(1000), uint64(2000), uint64(3000)]
    parent_puzhash_amounts = []
    merge_coin_ids: list[bytes32] = []
    for pid, amt in zip(parent_ids, amounts):
        parent_puzhash_amounts.append([pid, my_puzhash, amt])
        merge_coin_ids.append(Coin(pid, my_puzhash, amt).name())

    output_parent_amount = [output_parent_id, output_coin_amount]
    output_coin_id = Coin(output_parent_id, my_puzhash, output_coin_amount).name()

    agg_sol = Program.to([[output_coin_id, my_puzhash, output_coin_amount, parent_puzhash_amounts, 0]])
    agg_conds = conditions_dict_for_solution(p2_singleton, agg_sol, INFINITE_COST)
    # aggregator coin announces merge coin IDs
    agg_ccas = [std_hash(output_coin_id + x.vars[0]) for x in agg_conds[ConditionOpcode.CREATE_COIN_ANNOUNCEMENT]]
    # aggregator coin asserts 0 from merge coins
    agg_acas = [x.vars[0] for x in agg_conds[ConditionOpcode.ASSERT_COIN_ANNOUNCEMENT]]

    for coin_id, ppa in zip(merge_coin_ids, parent_puzhash_amounts):
        sol = Program.to([[coin_id, ppa[1], ppa[2], 0, output_parent_amount]])
        merge_conds = conditions_dict_for_solution(p2_singleton, sol, INFINITE_COST)
        # merge coin announces 0
        cca = std_hash(coin_id + merge_conds[ConditionOpcode.CREATE_COIN_ANNOUNCEMENT][0].vars[0])
        # merge coin asserts my_id from aggregator
        aca = merge_conds[ConditionOpcode.ASSERT_COIN_ANNOUNCEMENT][0].vars[0]
        assert aca in agg_ccas
        assert cca in agg_acas
        assert merge_conds[ConditionOpcode.ASSERT_MY_COIN_ID][0].vars[0] == coin_id


def test_treasury() -> None:
    """
    The treasury has two spend paths:
    - Proposal Path: when a proposal is being closed the treasury spend runs the
      validator and the actual proposed code (if passed)
    - Oracle Path: The treasury can make announcements about itself that are
      used to close invalid proposals
    """
    # Setup the treasury
    treasury_id = Program.to("treasury_id").get_tree_hash()
    treasury_struct: Program = Program.to((SINGLETON_MOD_HASH, (treasury_id, SINGLETON_LAUNCHER_HASH)))
    CAT_TAIL_HASH = Program.to("tail").get_tree_hash()

    proposal_id = Program.to("singleton_id").get_tree_hash()
    proposal_struct: Program = Program.to((SINGLETON_MOD_HASH, (proposal_id, SINGLETON_LAUNCHER_HASH)))
    p2_singleton = P2_SINGLETON_MOD.curry(treasury_struct, P2_SINGLETON_AGGREGATOR_MOD)
    p2_singleton_puzhash = p2_singleton.get_tree_hash()
    parent_id = Program.to("parent").get_tree_hash()
    locked_amount = 100000
    oracle_spend_delay = 10
    self_destruct_time = 1000
    proposal_length = 40
    soft_close_length = 5
    attendance = 1000
    pass_margin = 5100
    conditions = [[51, 0xDABBAD00, 1000], [51, 0xCAFEF00D, 100]]

    # Setup the validator
    minimum_amt = 1
    excess_puzhash = bytes32(b"1" * 32)
    dao_lockup_self = DAO_LOCKUP_MOD.curry(
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        DAO_FINISHED_STATE_HASH,
        CAT_MOD_HASH,
        CAT_TAIL_HASH,
    )

    proposal_curry_one = DAO_PROPOSAL_MOD.curry(
        DAO_PROPOSAL_TIMER_MOD_HASH,
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        CAT_MOD_HASH,
        DAO_FINISHED_STATE_HASH,
        DAO_TREASURY_MOD_HASH,
        dao_lockup_self.get_tree_hash(),
        CAT_TAIL_HASH,
        treasury_id,
    )
    proposal_validator = DAO_PROPOSAL_VALIDATOR_MOD.curry(
        treasury_struct,
        proposal_curry_one.get_tree_hash(),
        minimum_amt,
        excess_puzhash,
    )

    # Can now create the treasury inner puz
    treasury_inner = DAO_TREASURY_MOD.curry(
        DAO_TREASURY_MOD_HASH,
        proposal_validator,
        proposal_length,
        soft_close_length,
        attendance,
        pass_margin,
        self_destruct_time,
        oracle_spend_delay,
    )

    # Setup the spend_p2_singleton (proposal inner puz)
    spend_p2_singleton = SPEND_P2_SINGLETON_MOD.curry(
        treasury_struct,
        CAT_MOD_HASH,
        conditions,
        [],
        p2_singleton_puzhash,  # tailhash conds
    )
    spend_p2_singleton_puzhash = spend_p2_singleton.get_tree_hash()

    parent_amt_list = [[parent_id, locked_amount]]
    cat_parent_amt_list: list[Optional[Any]] = []
    spend_p2_singleton_solution = Program.to([parent_amt_list, cat_parent_amt_list, treasury_inner.get_tree_hash()])

    proposal: Program = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        proposal_id,
        spend_p2_singleton_puzhash,
        950,
        1200,
    )
    full_proposal = SINGLETON_MOD.curry(proposal_struct, proposal)

    # Oracle spend
    solution: Program = Program.to([0, 0, 0, 0, 0, treasury_struct])
    conds: Program = treasury_inner.run(solution)
    assert len(conds.as_python()) == 3

    # Proposal Spend
    proposal_amt = uint64(10)
    proposal_coin_id = Coin(parent_id, full_proposal.get_tree_hash(), proposal_amt).name()
    solution = Program.to(
        [
            [proposal_coin_id, spend_p2_singleton_puzhash, 0, "s"],
            [proposal_id, 1200, 950, parent_id, proposal_amt],
            spend_p2_singleton,
            spend_p2_singleton_solution,
        ]
    )
    conds = treasury_inner.run(solution)
    assert len(conds.as_python()) == 10 + len(conditions)


def test_lockup() -> None:
    """
    The lockup puzzle tracks the voting records of DAO CATs. When a proposal is
    voted on, the proposal ID is added to a list, against which, future votes
    are checked.
    This test checks the addition of new votes to the lockup, and that you can't
    re-vote on a proposal twice.
    """
    CAT_TAIL_HASH = Program.to("tail").get_tree_hash()

    INNERPUZ = Program.to(1)
    previous_votes = [0xFADEDDAB]

    dao_lockup_self = DAO_LOCKUP_MOD.curry(
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        DAO_FINISHED_STATE_HASH,
        CAT_MOD_HASH,
        CAT_TAIL_HASH,
    )

    full_lockup_puz: Program = dao_lockup_self.curry(
        dao_lockup_self.get_tree_hash(),
        previous_votes,
        INNERPUZ,
    )
    my_id = Program.to("my_id").get_tree_hash()
    lockup_coin_amount = 20

    # Test adding vote
    new_proposal = 0xBADDADAB
    new_vote_list = [new_proposal, 0xFADEDDAB]
    child_puzhash = dao_lockup_self.curry(
        dao_lockup_self.get_tree_hash(),
        new_vote_list,
        INNERPUZ,
    ).get_tree_hash()
    message = Program.to([new_proposal, lockup_coin_amount, 1, my_id]).get_tree_hash()
    generated_conditions = [[51, child_puzhash, lockup_coin_amount], [62, message]]
    solution: Program = Program.to(
        [
            my_id,
            generated_conditions,
            20,
            new_proposal,
            INNERPUZ.get_tree_hash(),  # fake proposal curry vals
            1,
            20,
            child_puzhash,
            0,
        ]
    )
    conds: Program = full_lockup_puz.run(solution)
    assert len(conds.as_python()) == 6

    # Test Re-voting on same proposal fails
    new_proposal = 0xBADDADAB
    new_vote_list = [new_proposal, 0xBADDADAB]
    child_puzhash = dao_lockup_self.curry(
        dao_lockup_self.get_tree_hash(),
        new_vote_list,
        INNERPUZ,
    ).get_tree_hash()
    message = Program.to([new_proposal, lockup_coin_amount, 1, my_id]).get_tree_hash()
    generated_conditions = [[51, child_puzhash, lockup_coin_amount], [62, message]]
    revote_solution: Program = Program.to(
        [
            my_id,
            generated_conditions,
            20,
            new_proposal,
            INNERPUZ.get_tree_hash(),  # fake proposal curry vals
            1,
            20,
            child_puzhash,
            0,
        ]
    )
    with pytest.raises(ValueError) as e_info:
        conds = full_lockup_puz.run(revote_solution)
    assert "klvm raise" in e_info.value.args[0]

    # Test vote removal
    solution = Program.to(
        [
            0,
            generated_conditions,
            20,
            [0xFADEDDAB],
            INNERPUZ.get_tree_hash(),
            0,
            0,
            0,
            0,
        ]
    )
    conds = full_lockup_puz.run(solution)
    assert len(conds.as_python()) == 3

    new_innerpuz = Program.to("new_inner")
    new_innerpuzhash = new_innerpuz.get_tree_hash()
    child_lockup = dao_lockup_self.curry(
        dao_lockup_self.get_tree_hash(),
        previous_votes,
        new_innerpuz,
    ).get_tree_hash()
    message = Program.to([0, 0, 0, my_id]).get_tree_hash()
    spend_conds = [[51, child_lockup, lockup_coin_amount], [62, message]]
    transfer_sol = Program.to(
        [
            my_id,
            spend_conds,
            lockup_coin_amount,
            0,
            INNERPUZ.get_tree_hash(),  # fake proposal curry vals
            0,
            0,
            INNERPUZ.get_tree_hash(),
            new_innerpuzhash,
        ]
    )
    conds = full_lockup_puz.run(transfer_sol)
    assert conds.at("rrrrfrf").as_atom() == child_lockup


def test_proposal_lifecycle() -> None:
    """
    This test covers the whole lifecycle of a proposal and treasury.
    Its main function is to check that the announcement pairs between treasury
    and proposal are accurate.
    It covers the spend proposal and update proposal types.
    """
    proposal_pass_percentage: uint64 = uint64(5100)
    attendance_required: uint64 = uint64(1000)
    proposal_timelock: uint64 = uint64(40)
    soft_close_length: uint64 = uint64(5)
    self_destruct_time: uint64 = uint64(1000)
    oracle_spend_delay: uint64 = uint64(10)
    min_amt: uint64 = uint64(1)
    CAT_TAIL_HASH = Program.to("tail").get_tree_hash()

    dao_rules = DAORules(
        proposal_timelock=proposal_timelock,
        soft_close_length=soft_close_length,
        attendance_required=attendance_required,
        pass_percentage=proposal_pass_percentage,
        self_destruct_length=self_destruct_time,
        oracle_spend_delay=oracle_spend_delay,
        proposal_minimum_amount=min_amt,
    )

    # Setup the treasury
    treasury_id = Program.to("treasury_id").get_tree_hash()
    treasury_singleton_struct: Program = Program.to((SINGLETON_MOD_HASH, (treasury_id, SINGLETON_LAUNCHER_HASH)))
    treasury_amount = 1

    # setup the p2_singleton
    p2_singleton = P2_SINGLETON_MOD.curry(treasury_singleton_struct, P2_SINGLETON_AGGREGATOR_MOD)
    p2_singleton_puzhash = p2_singleton.get_tree_hash()
    parent_id = Program.to("parent").get_tree_hash()
    locked_amount = 100000
    conditions = [[51, 0xDABBAD00, 1000], [51, 0xCAFEF00D, 100]]

    excess_puzhash = get_p2_singleton_puzhash(treasury_id)
    dao_lockup_self = DAO_LOCKUP_MOD.curry(
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        DAO_FINISHED_STATE_HASH,
        CAT_MOD_HASH,
        CAT_TAIL_HASH,
    )

    proposal_curry_one = DAO_PROPOSAL_MOD.curry(
        DAO_PROPOSAL_TIMER_MOD_HASH,
        SINGLETON_MOD_HASH,
        SINGLETON_LAUNCHER_HASH,
        CAT_MOD_HASH,
        DAO_FINISHED_STATE_HASH,
        DAO_TREASURY_MOD_HASH,
        dao_lockup_self.get_tree_hash(),
        CAT_TAIL_HASH,
        treasury_id,
    )
    proposal_validator = DAO_PROPOSAL_VALIDATOR_MOD.curry(
        treasury_singleton_struct,
        proposal_curry_one.get_tree_hash(),
        min_amt,
        excess_puzhash,
    )

    treasury_inner_puz: Program = DAO_TREASURY_MOD.curry(
        DAO_TREASURY_MOD_HASH,
        proposal_validator,
        proposal_timelock,
        soft_close_length,
        attendance_required,
        proposal_pass_percentage,
        self_destruct_time,
        oracle_spend_delay,
    )
    treasury_inner_puzhash = treasury_inner_puz.get_tree_hash()

    calculated_treasury_puzhash = get_treasury_puzzle(dao_rules, treasury_id, CAT_TAIL_HASH).get_tree_hash()
    assert treasury_inner_puzhash == calculated_treasury_puzhash

    full_treasury_puz = SINGLETON_MOD.curry(treasury_singleton_struct, treasury_inner_puz)
    full_treasury_puzhash = full_treasury_puz.get_tree_hash()

    # Setup the spend_p2_singleton (proposal inner puz)
    spend_p2_singleton = SPEND_P2_SINGLETON_MOD.curry(
        treasury_singleton_struct,
        CAT_MOD_HASH,
        conditions,
        [],
        p2_singleton_puzhash,  # tailhash conds
    )
    spend_p2_singleton_puzhash = spend_p2_singleton.get_tree_hash()

    parent_amt_list = [[parent_id, locked_amount]]
    cat_parent_amt_list: list[Optional[Any]] = []
    spend_p2_singleton_solution = Program.to([parent_amt_list, cat_parent_amt_list, treasury_inner_puzhash])

    # Setup Proposal
    proposal_id = Program.to("proposal_id").get_tree_hash()
    proposal_singleton_struct: Program = Program.to((SINGLETON_MOD_HASH, (proposal_id, SINGLETON_LAUNCHER_HASH)))

    current_votes = 1200
    yes_votes = 950
    proposal: Program = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        proposal_id,
        spend_p2_singleton_puzhash,
        yes_votes,
        current_votes,
    )
    full_proposal: Program = SINGLETON_MOD.curry(proposal_singleton_struct, proposal)
    full_proposal_puzhash: bytes32 = full_proposal.get_tree_hash()
    proposal_amt = uint64(11)
    proposal_coin_id = Coin(parent_id, full_proposal_puzhash, proposal_amt).name()

    treasury_solution: Program = Program.to(
        [
            [proposal_coin_id, spend_p2_singleton_puzhash, 0],
            [proposal_id, current_votes, yes_votes, parent_id, proposal_amt],
            spend_p2_singleton,
            spend_p2_singleton_solution,
        ]
    )

    proposal_solution = Program.to(
        [
            proposal_validator.get_tree_hash(),
            0,
            proposal_timelock,
            proposal_pass_percentage,
            attendance_required,
            0,
            soft_close_length,
            self_destruct_time,
            oracle_spend_delay,
            0,
            proposal_amt,
        ]
    )

    # lineage_proof my_amount inner_solution
    lineage_proof = [treasury_id, treasury_inner_puzhash, treasury_amount]
    full_treasury_solution = Program.to([lineage_proof, treasury_amount, treasury_solution])
    full_proposal_solution = Program.to([lineage_proof, proposal_amt, proposal_solution])

    # Run the puzzles
    treasury_conds = conditions_dict_for_solution(full_treasury_puz, full_treasury_solution, INFINITE_COST)
    proposal_conds = conditions_dict_for_solution(full_proposal, full_proposal_solution, INFINITE_COST)

    # Announcements
    treasury_aca = treasury_conds[ConditionOpcode.ASSERT_COIN_ANNOUNCEMENT][0].vars[0]
    proposal_cca = proposal_conds[ConditionOpcode.CREATE_COIN_ANNOUNCEMENT][0].vars[0]
    assert std_hash(proposal_coin_id + proposal_cca) == treasury_aca

    treasury_cpas = [
        std_hash(full_treasury_puzhash + cond.vars[0])
        for cond in treasury_conds[ConditionOpcode.CREATE_PUZZLE_ANNOUNCEMENT]
    ]
    proposal_apas = [cond.vars[0] for cond in proposal_conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT]]
    assert treasury_cpas[1] == proposal_apas[1]

    # Test Proposal to update treasury
    # Set up new treasury params
    new_proposal_pass_percentage: uint64 = uint64(2500)
    new_attendance_required: uint64 = uint64(500)
    new_proposal_timelock: uint64 = uint64(900)
    new_soft_close_length: uint64 = uint64(10)
    new_self_destruct_time: uint64 = uint64(1000)
    new_oracle_spend_delay: uint64 = uint64(20)
    new_minimum_amount: uint64 = uint64(10)
    proposal_excess_puzhash: bytes32 = get_p2_singleton_puzhash(treasury_id)

    new_dao_rules = DAORules(
        proposal_timelock=new_proposal_timelock,
        soft_close_length=new_soft_close_length,
        attendance_required=new_attendance_required,
        pass_percentage=new_proposal_pass_percentage,
        self_destruct_length=new_self_destruct_time,
        oracle_spend_delay=new_oracle_spend_delay,
        proposal_minimum_amount=new_minimum_amount,
    )

    update_proposal = DAO_UPDATE_MOD.curry(
        DAO_TREASURY_MOD_HASH,
        DAO_PROPOSAL_VALIDATOR_MOD_HASH,
        treasury_singleton_struct,
        proposal_curry_one.get_tree_hash(),
        new_minimum_amount,
        proposal_excess_puzhash,
        new_proposal_timelock,
        new_soft_close_length,
        new_attendance_required,
        new_proposal_pass_percentage,
        new_self_destruct_time,
        new_oracle_spend_delay,
    )
    update_proposal_puzhash = update_proposal.get_tree_hash()
    update_proposal_sol = Program.to([])

    proposal = proposal_curry_one.curry(
        proposal_curry_one.get_tree_hash(),
        proposal_id,
        update_proposal_puzhash,
        yes_votes,
        current_votes,
    )
    full_proposal = SINGLETON_MOD.curry(proposal_singleton_struct, proposal)
    full_proposal_puzhash = full_proposal.get_tree_hash()
    proposal_coin_id = Coin(parent_id, full_proposal_puzhash, proposal_amt).name()

    treasury_solution = Program.to(
        [
            [proposal_coin_id, update_proposal_puzhash, 0, "u"],
            [proposal_id, current_votes, yes_votes, parent_id, proposal_amt],
            update_proposal,
            update_proposal_sol,
        ]
    )

    proposal_solution = Program.to(
        [
            proposal_validator.get_tree_hash(),
            0,
            proposal_timelock,
            proposal_pass_percentage,
            attendance_required,
            0,
            soft_close_length,
            self_destruct_time,
            oracle_spend_delay,
            0,
            proposal_amt,
        ]
    )

    lineage_proof = [treasury_id, treasury_inner_puzhash, treasury_amount]
    full_treasury_solution = Program.to([lineage_proof, treasury_amount, treasury_solution])
    full_proposal_solution = Program.to([lineage_proof, proposal_amt, proposal_solution])

    treasury_conds = conditions_dict_for_solution(full_treasury_puz, full_treasury_solution, INFINITE_COST)
    proposal_conds = conditions_dict_for_solution(full_proposal, full_proposal_solution, INFINITE_COST)

    treasury_aca = treasury_conds[ConditionOpcode.ASSERT_COIN_ANNOUNCEMENT][0].vars[0]
    proposal_cca = proposal_conds[ConditionOpcode.CREATE_COIN_ANNOUNCEMENT][0].vars[0]
    assert std_hash(proposal_coin_id + proposal_cca) == treasury_aca

    treasury_cpas = [
        std_hash(full_treasury_puzhash + cond.vars[0])
        for cond in treasury_conds[ConditionOpcode.CREATE_PUZZLE_ANNOUNCEMENT]
    ]
    proposal_apas = [cond.vars[0] for cond in proposal_conds[ConditionOpcode.ASSERT_PUZZLE_ANNOUNCEMENT]]
    assert treasury_cpas[1] == proposal_apas[1]

    new_treasury_inner = update_proposal.run(update_proposal_sol).at("frf").as_atom()
    expected_treasury_inner = get_treasury_puzzle(new_dao_rules, treasury_id, CAT_TAIL_HASH)
    assert new_treasury_inner == expected_treasury_inner.get_tree_hash()

    expected_treasury_hash = curry_singleton(treasury_id, expected_treasury_inner).get_tree_hash()
    assert treasury_conds[ConditionOpcode.CREATE_COIN][1].vars[0] == expected_treasury_hash


async def do_spend(
    sim: SpendSim,
    sim_client: SimClient,
    coins: list[Coin],
    puzzles: list[Program],
    solutions: list[Program],
) -> tuple[MempoolInclusionStatus, Optional[Err]]:
    spends = []
    for coin, puzzle, solution in zip(coins, puzzles, solutions):
        spends.append(make_spend(coin, puzzle, solution))
    spend_bundle = WalletSpendBundle(spends, AugSchemeMPL.aggregate([]))
    result = await sim_client.push_tx(spend_bundle)
    await sim.farm_block()
    return result


@pytest.mark.anyio
async def test_singleton_aggregator() -> None:
    async with sim_and_client() as (sim, sim_client):
        aggregator = P2_SINGLETON_AGGREGATOR_MOD
        aggregator_hash = aggregator.get_tree_hash()
        await sim.farm_block(aggregator_hash)
        await sim.farm_block(aggregator_hash)
        for i in range(5):
            await sim.farm_block()

        coin_records = await sim_client.get_coin_records_by_puzzle_hash(aggregator_hash)
        coins = [c.coin for c in coin_records]

        output_coin = coins[0]
        output_sol = Program.to(
            [
                output_coin.name(),
                output_coin.puzzle_hash,
                output_coin.amount,
                [[c.parent_coin_info, c.puzzle_hash, c.amount] for c in coins[1:]],
            ]
        )
        merge_sols = [
            Program.to([c.name(), c.puzzle_hash, c.amount, [], [output_coin.parent_coin_info, output_coin.amount]])
            for c in coins[1:]
        ]

        res = await do_spend(sim, sim_client, coins, [aggregator] * 4, [output_sol, *merge_sols])
        assert res[0] == MempoolInclusionStatus.SUCCESS

        await sim.rewind(uint32(sim.block_height - 1))

        # Spend a merge coin with empty output details
        output_sol = Program.to(
            [
                output_coin.name(),
                output_coin.puzzle_hash,
                output_coin.amount,
                [],
                [],
            ]
        )
        res = await do_spend(sim, sim_client, [output_coin], [aggregator], [output_sol])
        assert res[0] == MempoolInclusionStatus.FAILED

        # Try to steal treasury coins with a phoney output
        acs = Program.to(1)
        acs_ph = acs.get_tree_hash()
        await sim.farm_block(acs_ph)
        bad_coin = (await sim_client.get_coin_records_by_puzzle_hash(acs_ph))[0].coin
        bad_sol = Program.to(
            [
                [ConditionOpcode.CREATE_COIN, acs_ph, sum(c.amount for c in coins)],
                *[[ConditionOpcode.ASSERT_COIN_ANNOUNCEMENT, std_hash(c.name())] for c in coins],
                [ConditionOpcode.CREATE_COIN_ANNOUNCEMENT, 0],
            ]
        )

        merge_sols = [
            Program.to([c.name(), c.puzzle_hash, c.amount, [], [bad_coin.parent_coin_info, bad_coin.amount]])
            for c in coins
        ]

        res = await do_spend(sim, sim_client, [bad_coin, *coins], [acs] + [aggregator] * 4, [bad_sol, *merge_sols])
        assert res[0] == MempoolInclusionStatus.FAILED



from __future__ import annotations

import pytest

from chik.protocols.shared_protocol import Capability
from chik.server.capabilities import known_active_capabilities
from chik.util.ints import uint16


@pytest.mark.parametrize(
    argnames=["values", "expected"],
    argvalues=[
        # nothing, not even Chik mainnet...
        [[], []],
        # single valid
        [[(uint16(Capability.BASE), "1")], [Capability.BASE]],
        # all capabilities
        [[(uint16(capability), "1") for capability in Capability], list(Capability)],
        # all capabilities plus some invalid
        [
            [
                *[(uint16(capability), "1") for capability in Capability],
                *[(uint16(max(Capability) + n), "1") for n in range(1, 10)],
            ],
            list(Capability),
        ],
        # all possible values
        [
            [(uint16(n), "1") for n in range(2**16)],
            list(Capability),
        ],
        # all possible invalid values
        [
            [(uint16(n), "1") for n in set(range(2**16)) - set(Capability)],
            [],
        ],
        # single invalid
        [[(uint16(max(Capability) + 1), "1")], []],
        # a few invalid
        [[(uint16(max(Capability) + n), "1") for n in range(1, 10)], []],
    ],
)
@pytest.mark.parametrize(
    argnames="duplicated",
    argvalues=[False, True],
    ids=lambda value: "duplicated" if value else "as-is",
)
@pytest.mark.parametrize(
    argnames="disabled",
    argvalues=[False, True],
    ids=lambda value: "disabled" if value else "enabled",
)
def test_known_active_capabilities_filter(
    values: list[tuple[uint16, str]],
    expected: list[Capability],
    duplicated: bool,
    disabled: bool,
) -> None:
    if duplicated:
        values *= 2

    if disabled:
        values = [(value, "0") for value, state in values]
        expected = []

    assert known_active_capabilities(values=values) == expected



from __future__ import annotations

import dataclasses
import logging
import operator
import sys
import time
from collections.abc import Awaitable
from math import ceil
from os import mkdir
from pathlib import Path
from shutil import copy
from typing import Any, Callable, Union, cast

import pytest

from chik._tests.conftest import HarvesterFarmerEnvironment
from chik._tests.plot_sync.test_delta import dummy_plot
from chik._tests.util.misc import assert_rpc_error
from chik._tests.util.rpc import validate_get_routes
from chik._tests.util.time_out_assert import time_out_assert, time_out_assert_custom_interval
from chik.consensus.coinbase import create_puzzlehash_for_pk
from chik.farmer.farmer import Farmer
from chik.plot_sync.receiver import Receiver, get_list_or_len
from chik.plotting.util import add_plot_directory
from chik.protocols import farmer_protocol
from chik.protocols.harvester_protocol import Plot
from chik.rpc.farmer_rpc_api import (
    FilterItem,
    PaginatedRequestData,
    PlotInfoRequestData,
    PlotPathRequestData,
    plot_matches_filter,
)
from chik.rpc.farmer_rpc_client import FarmerRpcClient
from chik.simulator.block_tools import get_plot_dir
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.bech32m import decode_puzzle_hash, encode_puzzle_hash
from chik.util.config import load_config, lock_and_load_config, save_config
from chik.util.hash import std_hash
from chik.util.ints import uint8, uint32, uint64
from chik.wallet.derive_keys import master_sk_to_wallet_sk, master_sk_to_wallet_sk_unhardened

log = logging.getLogger(__name__)


async def wait_for_plot_sync(receiver: Receiver, previous_last_sync_id: uint64) -> None:
    def wait() -> bool:
        current_last_sync_id = receiver.last_sync().sync_id
        return current_last_sync_id not in {0, previous_last_sync_id}

    await time_out_assert(30, wait)


async def wait_for_synced_receiver(farmer: Farmer, harvester_id: bytes32) -> None:
    def wait() -> bool:
        return (
            harvester_id in farmer.plot_sync_receivers and not farmer.plot_sync_receivers[harvester_id].initial_sync()
        )

    await time_out_assert(30, wait)


@pytest.mark.anyio
async def test_get_routes(harvester_farmer_environment: HarvesterFarmerEnvironment) -> None:
    farmer_service, farmer_rpc_client, harvester_service, harvester_rpc_client, _ = harvester_farmer_environment
    assert farmer_service.rpc_server is not None
    assert harvester_service.rpc_server is not None
    await validate_get_routes(farmer_rpc_client, farmer_service.rpc_server.rpc_api)
    await validate_get_routes(harvester_rpc_client, harvester_service.rpc_server.rpc_api)


@pytest.mark.parametrize("endpoint", ["get_harvesters", "get_harvesters_summary"])
@pytest.mark.anyio
async def test_farmer_get_harvesters_and_summary(
    harvester_farmer_environment: HarvesterFarmerEnvironment, endpoint: str
) -> None:
    _, farmer_rpc_client, harvester_service, harvester_rpc_client, _ = harvester_farmer_environment
    harvester = harvester_service._node

    harvester_plots = []

    async def non_zero_plots() -> bool:
        res = await harvester_rpc_client.get_plots()
        nonlocal harvester_plots
        harvester_plots = res["plots"]
        return len(harvester_plots) > 0

    await time_out_assert(10, non_zero_plots)

    async def test_get_harvesters() -> bool:
        nonlocal harvester_plots
        harvester.plot_manager.trigger_refresh()
        await time_out_assert(5, harvester.plot_manager.needs_refresh, value=False)
        farmer_res = await getattr(farmer_rpc_client, endpoint)()

        if len(list(farmer_res["harvesters"])) != 1:
            log.error(f"test_get_harvesters: invalid harvesters {list(farmer_res['harvesters'])}")
            return False

        if farmer_res["harvesters"][0]["last_sync_time"] is None:
            log.error(f"test_get_harvesters: sync not done {list(farmer_res['harvesters'])}")
            return False

        harvester_dict = farmer_res["harvesters"][0]
        counts_only: bool = endpoint == "get_harvesters_summary"

        if not counts_only:
            harvester_dict["plots"] = sorted(harvester_dict["plots"], key=lambda item: item["filename"])
            harvester_plots = sorted(harvester_plots, key=lambda item: item["filename"])

        assert harvester_dict["plots"] == get_list_or_len(harvester_plots, counts_only)
        assert harvester_dict["failed_to_open_filenames"] == get_list_or_len([], counts_only)
        assert harvester_dict["no_key_filenames"] == get_list_or_len([], counts_only)
        assert harvester_dict["duplicates"] == get_list_or_len([], counts_only)

        return True

    await time_out_assert_custom_interval(30, 1, test_get_harvesters)


@pytest.mark.anyio
async def test_farmer_signage_point_endpoints(harvester_farmer_environment: HarvesterFarmerEnvironment) -> None:
    farmer_service, farmer_rpc_client, _, _, _ = harvester_farmer_environment
    farmer_api = farmer_service._api

    assert (await farmer_rpc_client.get_signage_point(std_hash(b"2"))) is None
    assert len(await farmer_rpc_client.get_signage_points()) == 0

    async def have_signage_points() -> bool:
        return len(await farmer_rpc_client.get_signage_points()) > 0

    sp = farmer_protocol.NewSignagePoint(
        std_hash(b"1"), std_hash(b"2"), std_hash(b"3"), uint64(1), uint64(1000000), uint8(2), uint32(1)
    )
    await farmer_api.new_signage_point(sp)

    await time_out_assert(5, have_signage_points, True)
    assert (await farmer_rpc_client.get_signage_point(std_hash(b"2"))) is not None


@pytest.mark.anyio
async def test_farmer_reward_target_endpoints(harvester_farmer_environment: HarvesterFarmerEnvironment) -> None:
    farmer_service, farmer_rpc_client, _, _, bt = harvester_farmer_environment
    farmer_api = farmer_service._api

    targets_1 = await farmer_rpc_client.get_reward_targets(False)
    assert "have_pool_sk" not in targets_1
    assert "have_farmer_sk" not in targets_1
    targets_2 = await farmer_rpc_client.get_reward_targets(True, 2)
    assert targets_2["have_pool_sk"] and targets_2["have_farmer_sk"]

    new_ph: bytes32 = create_puzzlehash_for_pk(master_sk_to_wallet_sk(bt.farmer_master_sk, uint32(2)).get_g1())
    new_ph_2: bytes32 = create_puzzlehash_for_pk(master_sk_to_wallet_sk(bt.pool_master_sk, uint32(7)).get_g1())

    await farmer_rpc_client.set_reward_targets(encode_puzzle_hash(new_ph, "xck"), encode_puzzle_hash(new_ph_2, "xck"))
    targets_3 = await farmer_rpc_client.get_reward_targets(True, 10)
    assert decode_puzzle_hash(targets_3["farmer_target"]) == new_ph
    assert decode_puzzle_hash(targets_3["pool_target"]) == new_ph_2
    assert targets_3["have_pool_sk"] and targets_3["have_farmer_sk"]

    # limit the derivation search to 3 should fail to find the pool sk
    targets_4 = await farmer_rpc_client.get_reward_targets(True, 3)
    assert not targets_4["have_pool_sk"] and targets_4["have_farmer_sk"]

    # check observer addresses
    observer_farmer: bytes32 = create_puzzlehash_for_pk(
        master_sk_to_wallet_sk_unhardened(bt.farmer_master_sk, uint32(2)).get_g1()
    )
    observer_pool: bytes32 = create_puzzlehash_for_pk(
        master_sk_to_wallet_sk_unhardened(bt.pool_master_sk, uint32(7)).get_g1()
    )
    await farmer_rpc_client.set_reward_targets(
        encode_puzzle_hash(observer_farmer, "xck"), encode_puzzle_hash(observer_pool, "xck")
    )
    targets = await farmer_rpc_client.get_reward_targets(True, 10)
    assert decode_puzzle_hash(targets["farmer_target"]) == observer_farmer
    assert decode_puzzle_hash(targets["pool_target"]) == observer_pool
    assert targets["have_pool_sk"] and targets["have_farmer_sk"]

    root_path = farmer_api.farmer._root_path
    config = load_config(root_path, "config.yaml")
    assert config["farmer"]["xck_target_address"] == encode_puzzle_hash(observer_farmer, "xck")
    assert config["pool"]["xck_target_address"] == encode_puzzle_hash(observer_pool, "xck")

    new_ph_2_encoded = encode_puzzle_hash(new_ph_2, "xck")
    added_char = new_ph_2_encoded + "a"
    with pytest.raises(ValueError):
        await farmer_rpc_client.set_reward_targets(None, added_char)

    replaced_char = new_ph_2_encoded[0:-1] + "a"
    with pytest.raises(ValueError):
        await farmer_rpc_client.set_reward_targets(None, replaced_char)


@pytest.mark.anyio
async def test_farmer_get_pool_state(
    harvester_farmer_environment: HarvesterFarmerEnvironment, self_hostname: str
) -> None:
    farmer_service, farmer_rpc_client, _, _, _ = harvester_farmer_environment
    farmer_api = farmer_service._api

    assert len((await farmer_rpc_client.get_pool_state())["pool_state"]) == 0
    pool_list = [
        {
            "launcher_id": "ae4ef3b9bfe68949691281a015a9c16630fc8f66d48c19ca548fb80768791afa",
            "owner_public_key": "aa11e92274c0f6a2449fd0c7cfab4a38f943289dbe2214c808b36390c34eacfaa1d4c8f3c6ec582ac502ff32228679a0",  # noqa
            "payout_instructions": "c2b08e41d766da4116e388357ed957d04ad754623a915f3fd65188a8746cf3e8",
            "pool_url": self_hostname,
            "p2_singleton_puzzle_hash": "16e4bac26558d315cded63d4c5860e98deb447cc59146dd4de06ce7394b14f17",
            "target_puzzle_hash": "344587cf06a39db471d2cc027504e8688a0a67cce961253500c956c73603fd58",
        }
    ]

    root_path = farmer_api.farmer._root_path
    with lock_and_load_config(root_path, "config.yaml") as config:
        config["pool"]["pool_list"] = pool_list
        save_config(root_path, "config.yaml", config)
    await farmer_api.farmer.update_pool_state()

    pool_state = (await farmer_rpc_client.get_pool_state())["pool_state"]
    assert len(pool_state) == 1
    assert (
        pool_state[0]["pool_config"]["payout_instructions"]
        == "c2b08e41d766da4116e388357ed957d04ad754623a915f3fd65188a8746cf3e8"
    )
    await farmer_rpc_client.set_payout_instructions(
        bytes32.from_hexstr(pool_state[0]["pool_config"]["launcher_id"]), "1234vy"
    )
    await farmer_api.farmer.update_pool_state()
    pool_state = (await farmer_rpc_client.get_pool_state())["pool_state"]
    assert pool_state[0]["pool_config"]["payout_instructions"] == "1234vy"

    now = time.time()
    # Big arbitrary numbers used to be unlikely to accidentally collide.
    before_24h = (now - (25 * 60 * 60), 29984713)
    since_24h = (now - (23 * 60 * 60), 93049817)
    for p2_singleton_puzzle_hash, pool_dict in farmer_api.farmer.pool_state.items():
        for key in ["points_found_24h", "points_acknowledged_24h"]:
            pool_dict[key].insert(0, since_24h)
            pool_dict[key].insert(0, before_24h)

    sp = farmer_protocol.NewSignagePoint(
        std_hash(b"1"), std_hash(b"2"), std_hash(b"3"), uint64(1), uint64(1000000), uint8(2), uint32(1)
    )
    await farmer_api.new_signage_point(sp)
    client_pool_state = await farmer_rpc_client.get_pool_state()
    for pool_dict in client_pool_state["pool_state"]:
        for key in ["points_found_24h", "points_acknowledged_24h"]:
            assert pool_dict[key][0] == list(since_24h)


@pytest.mark.anyio
async def test_farmer_get_pool_state_plot_count(
    harvester_farmer_environment: HarvesterFarmerEnvironment, self_hostname: str
) -> None:
    farmer_service, farmer_rpc_client, harvester_service, _, _ = harvester_farmer_environment
    farmer_api = farmer_service._api

    async def wait_for_plot_sync() -> bool:
        try:
            farmer_res = await farmer_rpc_client.get_harvesters_summary()
            plots: int = farmer_res["harvesters"][0]["plots"]
            return plots > 0
        except Exception:
            return False

    await time_out_assert(15, wait_for_plot_sync, True)

    assert len((await farmer_rpc_client.get_pool_state())["pool_state"]) == 0

    pool_contract_puzzle_hash: bytes32 = bytes32.from_hexstr(
        "1b9d1eaa3c6a9b27cd90ad9070eb012794a74b277446417bc7b904145010c087"
    )
    pool_list = [
        {
            "launcher_id": "ae4ef3b9bfe68949691281a015a9c16630fc8f66d48c19ca548fb80768791afa",
            "owner_public_key": "aa11e92274c0f6a2449fd0c7cfab4a38f943289dbe2214c808b36390c34eacfaa1d4c8f3c6ec582ac502ff32228679a0",  # noqa
            "payout_instructions": "c2b08e41d766da4116e388357ed957d04ad754623a915f3fd65188a8746cf3e8",
            "pool_url": self_hostname,
            "p2_singleton_puzzle_hash": pool_contract_puzzle_hash.hex(),
            "target_puzzle_hash": "344587cf06a39db471d2cc027504e8688a0a67cce961253500c956c73603fd58",
        }
    ]

    root_path = farmer_api.farmer._root_path
    with lock_and_load_config(root_path, "config.yaml") as config:
        config["pool"]["pool_list"] = pool_list
        save_config(root_path, "config.yaml", config)
    await farmer_api.farmer.update_pool_state()

    pool_plot_count: int = (await farmer_rpc_client.get_pool_state())["pool_state"][0]["plot_count"]
    assert pool_plot_count == 5

    # TODO: Maybe improve this to not remove from Receiver directly but instead from the harvester and then wait for
    #       plot sync event.
    async def remove_all_and_validate() -> bool:
        nonlocal pool_plot_count
        receiver = farmer_api.farmer.plot_sync_receivers[harvester_service._server.node_id]
        for path, plot in receiver.plots().copy().items():
            if plot.pool_contract_puzzle_hash == pool_contract_puzzle_hash:
                del receiver.plots()[path]
                pool_plot_count -= 1
        plot_count: int = (await farmer_rpc_client.get_pool_state())["pool_state"][0]["plot_count"]
        assert plot_count == pool_plot_count
        return plot_count > 0

    await time_out_assert(15, remove_all_and_validate, False)
    assert (await farmer_rpc_client.get_pool_state())["pool_state"][0]["plot_count"] == 0


@pytest.mark.parametrize(
    "filter_item, match",
    [
        (FilterItem("filename", "1"), True),
        (FilterItem("filename", "12"), True),
        (FilterItem("filename", "123"), True),
        (FilterItem("filename", "1234"), False),
        (FilterItem("filename", "23"), True),
        (FilterItem("filename", "3"), True),
        (FilterItem("filename", "0123"), False),
        (FilterItem("pool_contract_puzzle_hash", None), True),
        (FilterItem("pool_contract_puzzle_hash", "1"), False),
    ],
)
def test_plot_matches_filter(filter_item: FilterItem, match: bool) -> None:
    assert plot_matches_filter(dummy_plot("123"), filter_item) == match


@pytest.mark.parametrize(
    "endpoint, filtering, sort_key, reverse, expected_plot_count",
    [
        (FarmerRpcClient.get_harvester_plots_valid, [], "filename", False, 20),
        (FarmerRpcClient.get_harvester_plots_valid, [], "size", True, 20),
        (
            FarmerRpcClient.get_harvester_plots_valid,
            [FilterItem("pool_contract_puzzle_hash", None)],
            "file_size",
            True,
            15,
        ),
        (
            FarmerRpcClient.get_harvester_plots_valid,
            [FilterItem("size", "20"), FilterItem("filename", "81")],
            "plot_id",
            False,
            4,
        ),
        (FarmerRpcClient.get_harvester_plots_invalid, [], None, True, 13),
        (FarmerRpcClient.get_harvester_plots_invalid, ["invalid_0"], None, False, 6),
        (FarmerRpcClient.get_harvester_plots_invalid, ["inval", "lid_1"], None, False, 2),
        (FarmerRpcClient.get_harvester_plots_keys_missing, [], None, True, 3),
        (FarmerRpcClient.get_harvester_plots_keys_missing, ["keys_missing_1"], None, False, 2),
        (FarmerRpcClient.get_harvester_plots_duplicates, [], None, True, 7),
        (FarmerRpcClient.get_harvester_plots_duplicates, ["duplicates_0"], None, False, 3),
    ],
)
@pytest.mark.anyio
@pytest.mark.skipif(sys.platform == "win32", reason="avoiding crashes on windows until we fix this (crashing workers)")
async def test_farmer_get_harvester_plots_endpoints(
    harvester_farmer_environment: HarvesterFarmerEnvironment,
    endpoint: Callable[[FarmerRpcClient, PaginatedRequestData], Awaitable[dict[str, Any]]],
    filtering: Union[list[FilterItem], list[str]],
    sort_key: str,
    reverse: bool,
    expected_plot_count: int,
) -> None:
    farmer_service, farmer_rpc_client, harvester_service, harvester_rpc_client, _ = harvester_farmer_environment
    harvester = harvester_service._node
    harvester_id = harvester_service._server.node_id
    receiver = farmer_service._api.farmer.plot_sync_receivers[harvester_id]

    if receiver.initial_sync():
        await wait_for_plot_sync(receiver, receiver.last_sync().sync_id)

    harvester_plots = (await harvester_rpc_client.get_plots())["plots"]
    plots = []

    request: PaginatedRequestData
    if endpoint == FarmerRpcClient.get_harvester_plots_valid:
        request = PlotInfoRequestData(
            harvester_id, uint32(0), uint32(0), cast(list[FilterItem], filtering), sort_key, reverse
        )
    else:
        request = PlotPathRequestData(harvester_id, uint32(0), uint32(0), cast(list[str], filtering), reverse)

    def add_plot_directories(prefix: str, count: int) -> list[Path]:
        new_paths = []
        for i in range(count):
            new_paths.append(harvester.root_path / f"{prefix}_{i}")
            mkdir(new_paths[-1])
            add_plot_directory(harvester.root_path, str(new_paths[-1]))
        return new_paths

    # Generate the plot data and
    if endpoint == FarmerRpcClient.get_harvester_plots_valid:
        plots = harvester_plots
    elif endpoint == FarmerRpcClient.get_harvester_plots_invalid:
        invalid_paths = add_plot_directories("invalid", 3)
        for dir_index, r in [(0, range(0, 6)), (1, range(6, 8)), (2, range(8, 13))]:
            plots += [str(invalid_paths[dir_index] / f"{i}.plot") for i in r]
        for plot in plots:
            with open(plot, "w"):
                pass
    elif endpoint == FarmerRpcClient.get_harvester_plots_keys_missing:
        keys_missing_plots = [path for path in (Path(get_plot_dir()) / "not_in_keychain").iterdir() if path.is_file()]
        keys_missing_paths = add_plot_directories("keys_missing", 2)
        for dir_index, copy_plots in [(0, keys_missing_plots[:1]), (1, keys_missing_plots[1:3])]:
            for plot in copy_plots:
                copy(plot, keys_missing_paths[dir_index])
                plots.append(str(keys_missing_paths[dir_index] / plot.name))

    elif endpoint == FarmerRpcClient.get_harvester_plots_duplicates:
        duplicate_paths = add_plot_directories("duplicates", 2)
        for dir_index, r in [(0, range(0, 3)), (1, range(3, 7))]:
            for i in r:
                plot_path = Path(harvester_plots[i]["filename"])
                plots.append(str(duplicate_paths[dir_index] / plot_path.name))
                copy(plot_path, plots[-1])

    # Sort and filter the data
    if endpoint == FarmerRpcClient.get_harvester_plots_valid:
        for filter_item in filtering:
            assert isinstance(filter_item, FilterItem)
            plots = [plot for plot in plots if plot_matches_filter(Plot.from_json_dict(plot), filter_item)]
        plots.sort(key=operator.itemgetter(sort_key, "plot_id"), reverse=reverse)
    else:
        for filter_item in filtering:
            plots = [plot for plot in plots if filter_item in plot]
        plots.sort(reverse=reverse)

    total_count = len(plots)
    assert total_count == expected_plot_count

    last_sync_id = receiver.last_sync().sync_id

    harvester.plot_manager.trigger_refresh()
    harvester.plot_manager.start_refreshing()

    await wait_for_plot_sync(receiver, last_sync_id)

    for page_size in [1, int(total_count / 2), total_count - 1, total_count, total_count + 1, 100]:
        request = dataclasses.replace(request, page_size=uint32(page_size))
        expected_page_count = ceil(total_count / page_size)
        for page in range(expected_page_count):
            request = dataclasses.replace(request, page=uint32(page))
            await wait_for_synced_receiver(farmer_service._api.farmer, harvester_id)
            page_result = await endpoint(farmer_rpc_client, request)
            offset = page * page_size
            expected_plots = plots[offset : offset + page_size]
            assert page_result == {
                "success": True,
                "node_id": harvester_id.hex(),
                "page": page,
                "page_count": expected_page_count,
                "total_count": total_count,
                "plots": expected_plots,
            }


@pytest.mark.anyio
@pytest.mark.skip("This test causes hangs occasionally. TODO: fix this.")
async def test_harvester_add_plot_directory(harvester_farmer_environment: HarvesterFarmerEnvironment) -> None:
    _, _, harvester_service, harvester_rpc_client, _ = harvester_farmer_environment

    async def assert_added(path: Path) -> None:
        assert await harvester_rpc_client.add_plot_directory(str(path))
        with lock_and_load_config(root_path, "config.yaml") as config:
            assert str(path) in config["harvester"]["plot_directories"]

    # Test without the required parameter: dirname
    with pytest.raises(ValueError, match="dirname"):
        await harvester_rpc_client.fetch("add_plot_directory", {})

    root_path = harvester_service._node.root_path
    test_path = Path(root_path / "test_path").resolve()

    # The test_path doesn't exist at this point
    with assert_rpc_error(f"Path doesn't exist: {test_path}"):
        await harvester_rpc_client.add_plot_directory(str(test_path))

    # Create a file at the test_path and make sure it detects this
    with open(test_path, "w"):
        pass

    with assert_rpc_error(f"Path is not a directory: {test_path}"):
        await harvester_rpc_client.add_plot_directory(str(test_path))

    # Drop the file, make it a directory and make sure it gets added properly.
    test_path.unlink()
    mkdir(test_path)

    await assert_added(test_path)

    with assert_rpc_error(f"Path already added: {test_path}"):
        await harvester_rpc_client.add_plot_directory(str(test_path))

    # Add another one and make sure they are still both there.
    test_path_other = test_path / "other"
    mkdir(test_path_other)
    await assert_added(test_path_other)

    added_directories = await harvester_rpc_client.get_plot_directories()
    assert str(test_path) in added_directories
    assert str(test_path_other) in added_directories



from __future__ import annotations

from pathlib import Path

from chik.simulator.block_tools import get_plot_dir


def get_test_plots(sub_dir: str = "") -> list[Path]:
    path = get_plot_dir()
    if sub_dir != "":
        path /= sub_dir
    return list(sorted(path.glob("*.plot")))



from __future__ import annotations

import pytest
from chikbip158 import PyBIP158


@pytest.mark.anyio
async def test_basic_filter_test(simulator_and_wallet):
    _full_nodes, wallets, bt = simulator_and_wallet
    wallet_node, _server_2 = wallets[0]
    wallet = wallet_node.wallet_state_manager.main_wallet

    num_blocks = 2
    ph = await wallet.get_new_puzzlehash()
    blocks = bt.get_consecutive_blocks(
        10,
        guarantee_transaction_block=True,
        farmer_reward_puzzle_hash=ph,
        pool_reward_puzzle_hash=ph,
    )
    for i in range(1, num_blocks):
        byte_array_tx: list[bytes] = []
        block = blocks[i]
        coins = block.get_included_reward_coins()
        coin_0 = bytearray(coins[0].puzzle_hash)
        coin_1 = bytearray(coins[1].puzzle_hash)
        byte_array_tx.append(coin_0)
        byte_array_tx.append(coin_1)

        pl = PyBIP158(byte_array_tx)
        present = pl.Match(coin_0)
        fee_present = pl.Match(coin_1)

        assert present
        assert fee_present



from __future__ import annotations

import random

import pytest

from chik.server.server import ChikServer
from chik.simulator.block_tools import BlockTools
from chik.simulator.full_node_simulator import FullNodeSimulator
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.peer_info import PeerInfo
from chik.util.ints import uint64
from chik.wallet.puzzles.clawback.metadata import ClawbackMetadata
from chik.wallet.wallet_node import WalletNode


@pytest.mark.parametrize(
    "trusted",
    [True, False],
)
@pytest.mark.anyio
async def test_is_recipient(
    simulator_and_wallet: tuple[list[FullNodeSimulator], list[tuple[WalletNode, ChikServer]], BlockTools],
    trusted: bool,
    self_hostname: str,
    seeded_random: random.Random,
) -> None:
    full_nodes, wallets, _ = simulator_and_wallet
    full_node_api = full_nodes[0]
    server_1: ChikServer = full_node_api.full_node.server
    wallet_node, server_2 = wallets[0]
    wallet = wallet_node.wallet_state_manager.main_wallet
    puzhash_1 = await wallet.get_new_puzzlehash()
    puzhash_2 = await wallet.get_new_puzzlehash()
    await server_2.start_client(PeerInfo(self_hostname, server_1.get_port()), None)
    invalid_data = ClawbackMetadata(uint64(500), bytes32.random(seeded_random), bytes32.random(seeded_random))
    both_data = ClawbackMetadata(uint64(500), puzhash_1, puzhash_2)
    sender_data = ClawbackMetadata(uint64(500), puzhash_1, bytes32.random(seeded_random))
    recipient_data = ClawbackMetadata(uint64(500), bytes32.random(seeded_random), puzhash_2)
    # Test invalid metadata
    has_exception = False
    try:
        await invalid_data.is_recipient(wallet_node.wallet_state_manager.puzzle_store)
    except ValueError:
        has_exception = True
    assert has_exception
    # Test valid metadata
    assert not (await both_data.is_recipient(wallet_node.wallet_state_manager.puzzle_store))
    assert await recipient_data.is_recipient(wallet_node.wallet_state_manager.puzzle_store)
    assert not (await sender_data.is_recipient(wallet_node.wallet_state_manager.puzzle_store))



from __future__ import annotations

from pathlib import Path
from shutil import rmtree

import pytest
from click.testing import CliRunner

from chik.cmds.chik import cli
from chik.util.config import lock_and_load_config, save_config
from chik.util.default_root import SIMULATOR_ROOT_PATH

mnemonic = (  # ignore any secret warnings
    "cup smoke miss park baby say island tomorrow segment lava bitter easily settle gift renew arrive kangaroo dilemma "
    "organ skin design salt history awesome"
)
fingerprint = 2640131813
std_farming_address = "txck1mh4qanzyawn3v4uphgaj2cg6hrjazwyp0sx653fhn9apg6mfajlqr0kd92"
burn_address = "txck1qqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqqm6kslrezru"  # 0x0...dead

SIMULATOR_ROOT_PATH.mkdir(parents=True, exist_ok=True)  # this simplifies code later


def get_profile_path(starting_string: str) -> str:
    """
    Returns the name of a profile that does not exist yet.
    """
    i = 0
    while Path(SIMULATOR_ROOT_PATH / (starting_string + str(i))).exists():
        i += 1
    return starting_string + str(i)


@pytest.mark.skip("TODO: Fix this - hangs often in CI - root cause unknown")
def test_every_simulator_command() -> None:
    starting_str = "ci_test"
    simulator_name = get_profile_path(starting_str)
    runner: CliRunner = CliRunner()
    address = std_farming_address
    start_result = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "create", "-bm", mnemonic],
        catch_exceptions=False,
    )
    assert start_result.exit_code == 0
    assert f"Farming & Prefarm reward address: {address}" in start_result.output
    assert "chik_full_node_simulator: started" in start_result.output
    assert "Genesis block generated, exiting." in start_result.output

    config_dir = SIMULATOR_ROOT_PATH.joinpath(simulator_name)
    with lock_and_load_config(config_dir, "config.yaml") as config:
        config["rpc_timeout"] = 600
        save_config(config_dir, "config.yaml", config)

    try:
        # run all tests
        run_all_tests(runner, address, simulator_name)
    finally:
        stop_simulator(runner, simulator_name)


@pytest.mark.skip("Need to rewrite")
def test_custom_farming_address() -> None:
    runner: CliRunner = CliRunner()
    address = burn_address
    starting_str = "ci_address_test"
    simulator_name = get_profile_path(starting_str)
    start_result = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "create", "-bm", mnemonic, "--reward-address", address],
        catch_exceptions=False,
    )
    assert start_result.exit_code == 0
    assert f"Farming & Prefarm reward address: {address}" in start_result.output
    assert "chik_full_node_simulator: started" in start_result.output
    assert "Genesis block generated, exiting." in start_result.output

    try:
        # just run status test
        _test_sim_status(runner, address, simulator_name)
    finally:
        stop_simulator(runner, simulator_name)


def stop_simulator(runner: CliRunner, simulator_name: str) -> None:
    """Stop simulator."""
    result = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "stop", "-d"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert "chik_full_node_simulator: Stopped\nDaemon stopped\n" == result.output
    rmtree(SIMULATOR_ROOT_PATH / simulator_name)


def run_all_tests(runner: CliRunner, address: str, simulator_name: str) -> None:
    """Run all tests."""
    _test_sim_status(runner, address, simulator_name)
    _test_farm_and_revert_block(runner, address, simulator_name)


def _test_sim_status(runner: CliRunner, address: str, simulator_name: str) -> None:
    # show everything
    result = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "status", "--show-key", "-cia"],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    # asserts are grouped by arg
    assert (
        f"Fingerprint: {fingerprint}" in result.output
        and f"Mnemonic seed (24 secret words):\n{mnemonic}" in result.output
    )  # -k

    assert (
        "Network: simulator0" in result.output and "Current Blockchain Status: Full Node Synced" in result.output
    )  # default
    assert "Height:          1" in result.output  # default
    assert f"Current Farming address: {address}, with a balance of: 21000000.0 TXCK." in result.output  # default

    assert (
        f"Address: {address} has a balance of: 21000000000000000000 mojo, with a total of: 2 transactions."
        in result.output
    )  # -ia
    assert "Coin Amount: 2625000000000000000 mojo" in result.output  # -ic


def _test_farm_and_revert_block(runner: CliRunner, address: str, simulator_name: str) -> None:
    # make 5 blocks
    five_blocks_result = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "farm", "-b", "5", "-a", address],
        catch_exceptions=False,
    )
    assert five_blocks_result.exit_code == 0
    assert "Farmed 5 Transaction blocks" in five_blocks_result.output

    # check that height increased
    five_blocks_check = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "status"],
        catch_exceptions=False,
    )
    assert five_blocks_check.exit_code == 0
    assert "Height:          6" in five_blocks_check.output

    # do a reorg, 3 blocks back, 2 blocks forward, height now 8
    reorg_result = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "revert", "-b", "3", "-n", "2"],
        catch_exceptions=False,
    )
    assert reorg_result.exit_code == 0
    assert "Block: 3 and above " in reorg_result.output and "Block Height is now: 8" in reorg_result.output

    # check that height changed by 2
    reorg_check = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "status"],
        catch_exceptions=False,
    )
    assert reorg_check.exit_code == 0
    assert "Height:          8" in reorg_check.output

    # do a forceful reorg 4 blocks back
    forced_reorg_result = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "revert", "-b", "4", "-fd"],
        catch_exceptions=False,
    )
    assert forced_reorg_result.exit_code == 0
    assert (
        "Block: 8 and above were successfully deleted" in forced_reorg_result.output
        and "Block Height is now: 4" in forced_reorg_result.output
    )

    # check that height changed by 4
    forced_reorg_check = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "status"],
        catch_exceptions=False,
    )
    assert forced_reorg_check.exit_code == 0
    assert "Height:          4" in forced_reorg_check.output

    # test chain reset to genesis
    genesis_reset_result = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "revert", "-fd", "--reset"],
        catch_exceptions=False,
    )
    assert genesis_reset_result.exit_code == 0
    assert (
        "Block: 2 and above were successfully deleted" in genesis_reset_result.output
        and "Block Height is now: 1" in genesis_reset_result.output
    )

    # check that height changed to 1
    genesis_reset_check = runner.invoke(
        cli,
        ["dev", "sim", "-n", simulator_name, "status"],
        catch_exceptions=False,
    )
    assert genesis_reset_check.exit_code == 0
    assert "Height:          1" in genesis_reset_check.output



from __future__ import annotations

import asyncio
import logging
import socket
import ssl
import struct
import sys

if sys.platform == "win32":
    import _overlapped
    import _winapi

from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

from typing_extensions import Protocol, TypeAlias

# https://github.com/python/asyncio/pull/448
global_max_concurrent_connections: int = 250


if TYPE_CHECKING:
    # typeshed, for mypy, doesn't include the private attributes.  Documenting them
    # here so the actual inheriting code can be left tidy.

    # https://github.com/python/typeshed/blob/d084079fc3d89a7b51b89095ad67762944e0ace3/stdlib/asyncio/base_events.pyi#L22-L25
    # _ProtocolFactory: TypeAlias = Callable[[], asyncio.protocols.BaseProtocol]
    class _ProtocolFactory(Protocol):
        # https://github.com/python/mypy/issues/6910#issuecomment-1081107831
        # https://github.com/python/typeshed/pull/5718/files
        def __call__(self) -> asyncio.protocols.BaseProtocol: ...

    _SSLContext: TypeAlias = Union[bool, ssl.SSLContext, None]

    # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/base_events.py#L389
    # https://github.com/python/typeshed/blob/d084079fc3d89a7b51b89095ad67762944e0ace3/stdlib/asyncio/base_events.pyi#L64
    class EventsAbstractEventLoop(asyncio.events.AbstractEventLoop):
        # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/selector_events.py#L142
        # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/proactor_events.py#L814
        def _start_serving(
            self,
            protocol_factory: _ProtocolFactory,
            sock: socket.socket,
            sslcontext: Optional[_SSLContext] = ...,
            server: Optional[asyncio.base_events.Server] = ...,
            backlog: int = ...,
            # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/constants.py#L16
            ssl_handshake_timeout: Optional[float] = ...,
        ) -> None: ...

    # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/base_events.py#L278
    # https://github.com/python/typeshed/blob/d084079fc3d89a7b51b89095ad67762944e0ace3/stdlib/asyncio/base_events.pyi#L27
    class BaseEventsServer(asyncio.base_events.Server):
        if sys.platform == "win32":
            _loop: ChikProactorEventLoop
        else:
            _loop: EventsAbstractEventLoop
        _sockets: Iterable[socket.socket]
        _active_count: int
        _protocol_factory: _ProtocolFactory
        _backlog: int
        _ssl_context: _SSLContext
        _ssl_handshake_timeout: Optional[float]

        def _attach(self) -> None: ...

        def _detach(self) -> None: ...

        def _start_serving(self) -> None: ...

    if sys.platform == "win32":
        # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/windows_events.py#L48
        class _OverlappedFuture(asyncio.Future[Any]): ...

        # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/windows_events.py#L410
        # https://github.com/python/typeshed/blob/d084079fc3d89a7b51b89095ad67762944e0ace3/stdlib/asyncio/windows_events.pyi#L44
        class IocpProactor(asyncio.windows_events.IocpProactor):
            _loop: Optional[asyncio.events.AbstractEventLoop]

            def _register_with_iocp(self, obj: object) -> None: ...

            def _register(
                self,
                ov: _overlapped.Overlapped,
                obj: socket.socket,
                callback: Callable[[object, socket.socket, _overlapped.Overlapped], tuple[socket.socket, object]],
            ) -> _OverlappedFuture: ...

            def _get_accept_socket(self, family: socket.AddressFamily) -> socket.socket: ...

        # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/windows_events.py#L309
        # https://github.com/python/typeshed/blob/d084079fc3d89a7b51b89095ad67762944e0ace3/stdlib/asyncio/windows_events.pyi#L35
        class ProactorEventLoop(asyncio.windows_events.ProactorEventLoop):
            # Actually provided on BaseProactorEventLoop
            # https://github.com/python/cpython/blob/v3.10.8/Lib/asyncio/proactor_events.py#L627
            # https://github.com/python/typeshed/blob/d084079fc3d89a7b51b89095ad67762944e0ace3/stdlib/asyncio/proactor_events.pyi#L75
            _proactor: Any

else:
    BaseEventsServer = asyncio.base_events.Server
    if sys.platform == "win32":
        IocpProactor = asyncio.windows_events.IocpProactor
        ProactorEventLoop = asyncio.windows_events.ProactorEventLoop


class PausableServer(BaseEventsServer):
    _paused: bool
    max_concurrent_connections: int

    # https://github.com/python/typeshed/blob/d084079fc3d89a7b51b89095ad67762944e0ace3/stdlib/asyncio/base_events.pyi#L40-L48
    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        sockets: Iterable[socket.socket],
        protocol_factory: _ProtocolFactory,
        ssl_context: _SSLContext,
        backlog: int,
        ssl_handshake_timeout: Optional[float],
        max_concurrent_connections: Optional[int] = None,
    ) -> None:
        super().__init__(
            loop=loop,
            sockets=sockets,
            protocol_factory=protocol_factory,
            ssl_context=ssl_context,
            backlog=backlog,
            ssl_handshake_timeout=ssl_handshake_timeout,
        )
        self._paused = False
        self.max_concurrent_connections = (
            max_concurrent_connections if max_concurrent_connections is not None else global_max_concurrent_connections
        )

    def _attach(self) -> None:
        super()._attach()
        logging.getLogger(__name__).debug(f"New connection. Total connections: {self._active_count}")
        if not self._paused and self._active_count >= self.max_concurrent_connections:
            self._chik_pause()

    def _detach(self) -> None:
        super()._detach()
        logging.getLogger(__name__).debug(f"Connection lost. Total connections: {self._active_count}")
        if (
            self._active_count > 0
            and self._sockets is not None
            and self._paused
            and self._active_count < self.max_concurrent_connections
        ):
            self._chik_resume()

    def _chik_pause(self) -> None:
        """Pause future calls to accept()."""
        self._paused = True
        if sys.platform == "win32":
            # proactor
            self._loop.disable_connections()
        else:
            # selector
            for sock in self._sockets:
                self._loop.remove_reader(sock.fileno())
        logging.getLogger(__name__).debug("Maximum number of connections reached, paused accepting connections.")

    def _chik_resume(self) -> None:
        """Resume use of accept() on listening socket(s)."""
        self._paused = False
        if sys.platform == "win32":
            # proactor
            self._loop.enable_connections()
        else:
            # selector
            for sock in self._sockets:
                self._loop._start_serving(
                    self._protocol_factory,
                    sock,
                    self._ssl_context,
                    self,
                    self._backlog,
                    self._ssl_handshake_timeout,
                )
        logging.getLogger(__name__).debug("Resumed accepting connections.")


async def _chik_create_server(
    cls: Any,
    protocol_factory: _ProtocolFactory,
    host: Any,
    port: Any,
    *,
    family: socket.AddressFamily = socket.AF_UNSPEC,
    flags: socket.AddressInfo = socket.AI_PASSIVE,
    sock: Any = None,
    backlog: int = 100,
    ssl: _SSLContext = None,
    reuse_address: Optional[bool] = None,
    reuse_port: Optional[bool] = None,
    ssl_handshake_timeout: Optional[float] = 30,
    start_serving: bool = True,
) -> PausableServer:
    server: BaseEventsServer = await cls.create_server(
        protocol_factory=protocol_factory,
        host=host,
        port=port,
        family=family,
        flags=flags,
        sock=sock,
        backlog=backlog,
        ssl=ssl,
        reuse_address=reuse_address,
        reuse_port=reuse_port,
        # TODO: tweaked
        ssl_handshake_timeout=ssl_handshake_timeout if ssl is not None else None,
        start_serving=False,
    )
    pausable_server = PausableServer(
        loop=server._loop,
        sockets=server._sockets,
        protocol_factory=server._protocol_factory,
        ssl_context=server._ssl_context,
        backlog=server._backlog,
        ssl_handshake_timeout=server._ssl_handshake_timeout,
    )
    if start_serving:
        pausable_server._start_serving()
        # Skip one loop iteration so that all 'loop.add_reader'
        # go through.
        await asyncio.sleep(0)

    return pausable_server


class ChikSelectorEventLoop(asyncio.SelectorEventLoop):
    # Ignoring lack of typing since we are passing through and also never call this
    # ourselves. There may be a good solution if needed in the future.  We should
    # generally get a warning about calling an untyped function in case we do.
    async def create_server(self, *args, **kwargs) -> PausableServer:  # type: ignore[no-untyped-def]
        return await _chik_create_server(super(), *args, **kwargs)


if sys.platform == "win32":

    class ChikProactor(IocpProactor):
        allow_connections: bool

        def __init__(self, *args: Any, **kwargs: Any) -> None:
            super().__init__(*args, **kwargs)
            self.allow_connections = True

        def enable_connections(self) -> None:
            self.allow_connections = True

        def disable_connections(self) -> None:
            self.allow_connections = False

        async def _chik_accept_loop(self, listener: socket.socket) -> tuple[socket.socket, tuple[object, ...]]:
            while True:
                # TODO: switch to event drive code
                while not self.allow_connections:  # noqa: ASYNC110
                    await asyncio.sleep(0.01)

                try:
                    return await self._chik_accept(listener)
                except OSError as exc:
                    if exc.winerror not in {
                        _winapi.ERROR_NETNAME_DELETED,
                        _winapi.ERROR_OPERATION_ABORTED,
                    }:
                        raise

        def _chik_accept(self, listener: socket.socket) -> asyncio.Future[tuple[socket.socket, tuple[object, ...]]]:
            self._register_with_iocp(listener)
            conn = self._get_accept_socket(listener.family)
            ov = _overlapped.Overlapped(_winapi.NULL)
            ov.AcceptEx(listener.fileno(), conn.fileno())

            def finish_accept(
                trans: object, key: socket.socket, ov: _overlapped.Overlapped
            ) -> tuple[socket.socket, object]:
                ov.getresult()
                # Use SO_UPDATE_ACCEPT_CONTEXT so getsockname() etc work.
                buf = struct.pack("@P", listener.fileno())
                conn.setsockopt(socket.SOL_SOCKET, _overlapped.SO_UPDATE_ACCEPT_CONTEXT, buf)
                conn.settimeout(listener.gettimeout())
                return conn, conn.getpeername()

            async def accept_coro(self: ChikProactor, future: asyncio.Future[object], conn: socket.socket) -> None:
                # Coroutine closing the accept socket if the future is cancelled
                try:
                    await future
                except asyncio.CancelledError:
                    conn.close()
                    raise
                except OSError as exc:
                    # https://github.com/python/cpython/issues/93821#issuecomment-1157945855
                    if exc.winerror not in {
                        _winapi.ERROR_NETNAME_DELETED,
                        _winapi.ERROR_OPERATION_ABORTED,
                    }:
                        raise

            future = self._register(ov, listener, finish_accept)
            coro = accept_coro(self, future, conn)
            asyncio.ensure_future(coro, loop=self._loop)  # noqa: RUF006
            return future

        def accept(self, listener: socket.socket) -> asyncio.Future[tuple[socket.socket, tuple[object, ...]]]:
            coro = self._chik_accept_loop(listener)
            return asyncio.ensure_future(coro)

    class ChikProactorEventLoop(ProactorEventLoop):
        # Ignoring lack of typing (via Any) since we are passing through and also never
        # call this ourselves.  There may be a good solution if needed in the future.
        # It would be better to use a real ignore since then we would get a complaint
        # if we were to start calling this but we can not do that since this is
        # platform dependent code and running mypy on other platforms will complain
        # about the ignore being unused.
        async def create_server(self, *args: Any, **kwargs: Any) -> PausableServer:
            return await _chik_create_server(super(), *args, **kwargs)

        def __init__(self) -> None:
            proactor = ChikProactor()
            super().__init__(proactor=proactor)

        def enable_connections(self) -> None:
            self._proactor.enable_connections()

        def disable_connections(self) -> None:
            self._proactor.disable_connections()


class ChikPolicy(asyncio.DefaultEventLoopPolicy):
    def new_event_loop(self) -> asyncio.AbstractEventLoop:
        # overriding https://github.com/python/cpython/blob/v3.11.0/Lib/asyncio/events.py#L689-L695
        if sys.platform == "win32":
            loop_factory = ChikProactorEventLoop
        else:
            loop_factory = ChikSelectorEventLoop

        return loop_factory()


def set_chik_policy(connection_limit: int) -> None:
    global global_max_concurrent_connections
    # Add "+100" to the desired peer count as a safety margin.
    global_max_concurrent_connections = connection_limit + 100
    asyncio.set_event_loop_policy(ChikPolicy())



from __future__ import annotations

from enum import Enum


class ProtocolMessageTypes(Enum):
    # Shared protocol (all services)
    handshake = 1

    # Harvester protocol (harvester <-> farmer)
    harvester_handshake = 3
    # new_signage_point_harvester = 4 Changed to 66 in new protocol
    new_proof_of_space = 5
    request_signatures = 6
    respond_signatures = 7

    # Farmer protocol (farmer <-> full_node)
    new_signage_point = 8
    declare_proof_of_space = 9
    request_signed_values = 10
    signed_values = 11
    farming_info = 12

    # Timelord protocol (timelord <-> full_node)
    new_peak_timelord = 13
    new_unfinished_block_timelord = 14
    new_infusion_point_vdf = 15
    new_signage_point_vdf = 16
    new_end_of_sub_slot_vdf = 17
    request_compact_proof_of_time = 18
    respond_compact_proof_of_time = 19

    # Full node protocol (full_node <-> full_node)
    new_peak = 20
    new_transaction = 21
    request_transaction = 22
    respond_transaction = 23
    request_proof_of_weight = 24
    respond_proof_of_weight = 25
    request_block = 26
    respond_block = 27
    reject_block = 28
    request_blocks = 29
    respond_blocks = 30
    reject_blocks = 31
    new_unfinished_block = 32
    request_unfinished_block = 33
    respond_unfinished_block = 34
    new_signage_point_or_end_of_sub_slot = 35
    request_signage_point_or_end_of_sub_slot = 36
    respond_signage_point = 37
    respond_end_of_sub_slot = 38
    request_mempool_transactions = 39
    request_compact_vdf = 40
    respond_compact_vdf = 41
    new_compact_vdf = 42
    request_peers = 43
    respond_peers = 44
    none_response = 91

    # Wallet protocol (wallet <-> full_node)
    request_puzzle_solution = 45
    respond_puzzle_solution = 46
    reject_puzzle_solution = 47
    send_transaction = 48
    transaction_ack = 49
    new_peak_wallet = 50
    request_block_header = 51
    respond_block_header = 52
    reject_header_request = 53
    request_removals = 54
    respond_removals = 55
    reject_removals_request = 56
    request_additions = 57
    respond_additions = 58
    reject_additions_request = 59
    request_header_blocks = 60
    reject_header_blocks = 61
    respond_header_blocks = 62

    # Introducer protocol (introducer <-> full_node)
    request_peers_introducer = 63
    respond_peers_introducer = 64

    # Simulator protocol
    farm_new_block = 65

    # New harvester protocol
    new_signage_point_harvester = 66
    request_plots = 67
    respond_plots = 68
    plot_sync_start = 78
    plot_sync_loaded = 79
    plot_sync_removed = 80
    plot_sync_invalid = 81
    plot_sync_keys_missing = 82
    plot_sync_duplicates = 83
    plot_sync_done = 84
    plot_sync_response = 85

    # More wallet protocol
    coin_state_update = 69
    register_for_ph_updates = 70
    respond_to_ph_updates = 71
    register_for_coin_updates = 72
    respond_to_coin_updates = 73
    request_children = 74
    respond_children = 75
    request_ses_hashes = 76
    respond_ses_hashes = 77
    request_block_headers = 86
    reject_block_headers = 87
    respond_block_headers = 88
    request_fee_estimates = 89
    respond_fee_estimates = 90

    # new Full Node protocol messages
    new_unfinished_block2 = 92
    request_unfinished_block2 = 93

    # New wallet sync protocol
    request_remove_puzzle_subscriptions = 94
    respond_remove_puzzle_subscriptions = 95
    request_remove_coin_subscriptions = 96
    respond_remove_coin_subscriptions = 97
    request_puzzle_state = 98
    respond_puzzle_state = 99
    reject_puzzle_state = 100
    request_coin_state = 101
    respond_coin_state = 102
    reject_coin_state = 103

    # Wallet protocol mempool updates
    mempool_items_added = 104
    mempool_items_removed = 105
    request_cost_info = 106
    respond_cost_info = 107

    error = 255



# Package: utils

# Copyright (c) 2017 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import annotations

from collections.abc import Iterable
from typing import Optional

from chik_rs.sized_bytes import bytes32

# Based on this specification from Pieter Wuille:
# https://github.com/sipa/bips/blob/bip-bech32m/bip-bech32m.mediawiki
"""Reference implementation for Bech32m and segwit addresses."""


CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"


def bech32_polymod(values: list[int]) -> int:
    """Internal function that computes the Bech32 checksum."""
    generator = [0x3B6A57B2, 0x26508E6D, 0x1EA119FA, 0x3D4233DD, 0x2A1462B3]
    chk = 1
    for value in values:
        top = chk >> 25
        chk = (chk & 0x1FFFFFF) << 5 ^ value
        for i in range(5):
            chk ^= generator[i] if ((top >> i) & 1) else 0
    return chk


def bech32_hrp_expand(hrp: str) -> list[int]:
    """Expand the HRP into values for checksum computation."""
    return [ord(x) >> 5 for x in hrp] + [0] + [ord(x) & 31 for x in hrp]


M = 0x2BC830A3


def bech32_verify_checksum(hrp: str, data: list[int]) -> bool:
    return bech32_polymod(bech32_hrp_expand(hrp) + data) == M


def bech32_create_checksum(hrp: str, data: list[int]) -> list[int]:
    values = bech32_hrp_expand(hrp) + data
    polymod = bech32_polymod([*values, 0, 0, 0, 0, 0, 0]) ^ M
    return [(polymod >> 5 * (5 - i)) & 31 for i in range(6)]


def bech32_encode(hrp: str, data: list[int]) -> str:
    """Compute a Bech32 string given HRP and data values."""
    combined = data + bech32_create_checksum(hrp, data)
    return hrp + "1" + "".join([CHARSET[d] for d in combined])


def bech32_decode(bech: str, max_length: int = 90) -> tuple[Optional[str], Optional[list[int]]]:
    """Validate a Bech32 string, and determine HRP and data."""
    bech = bech.strip()
    if (any(ord(x) < 33 or ord(x) > 126 for x in bech)) or (bech.lower() != bech and bech.upper() != bech):
        return (None, None)
    bech = bech.lower()
    pos = bech.rfind("1")
    if pos < 1 or pos + 7 > len(bech) or len(bech) > max_length:
        return (None, None)
    if not all(x in CHARSET for x in bech[pos + 1 :]):
        return (None, None)
    hrp = bech[:pos]
    data = [CHARSET.find(x) for x in bech[pos + 1 :]]
    if not bech32_verify_checksum(hrp, data):
        return (None, None)
    return hrp, data[:-6]


def convertbits(data: Iterable[int], frombits: int, tobits: int, pad: bool = True) -> list[int]:
    """General power-of-2 base conversion."""
    acc = 0
    bits = 0
    ret = []
    maxv = (1 << tobits) - 1
    max_acc = (1 << (frombits + tobits - 1)) - 1
    for value in data:
        if value < 0 or (value >> frombits):
            raise ValueError("Invalid Value")
        acc = ((acc << frombits) | value) & max_acc
        bits += frombits
        while bits >= tobits:
            bits -= tobits
            ret.append((acc >> bits) & maxv)
    if pad:
        if bits:
            ret.append((acc << (tobits - bits)) & maxv)
    elif bits >= frombits or ((acc << (tobits - bits)) & maxv):
        raise ValueError("Invalid bits")
    return ret


def encode_puzzle_hash(puzzle_hash: bytes32, prefix: str) -> str:
    encoded = bech32_encode(prefix, convertbits(puzzle_hash, 8, 5))
    return encoded


def decode_puzzle_hash(address: str) -> bytes32:
    _hrpgot, data = bech32_decode(address)
    if data is None:
        raise ValueError("Invalid Address")
    decoded = convertbits(data, 5, 8, False)
    decoded_bytes = bytes32(decoded)
    return decoded_bytes



from __future__ import annotations

import random

import pytest

from chik.full_node.sync_store import SyncStore
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.util.hash import std_hash


@pytest.mark.anyio
async def test_basic_store():
    store = SyncStore()

    # Save/get sync
    for sync_mode in (False, True):
        store.set_sync_mode(sync_mode)
        assert sync_mode == store.get_sync_mode()

    peer_ids = [std_hash(bytes([a])) for a in range(3)]

    assert store.get_peers_that_have_peak([]) == set()
    assert store.get_peers_that_have_peak([std_hash(b"block1")]) == set()

    assert store.get_heaviest_peak() is None
    assert len(store.get_peak_of_each_peer()) == 0
    store.peer_has_block(std_hash(b"block10"), peer_ids[0], 500, 10, True)
    store.peer_has_block(std_hash(b"block1"), peer_ids[0], 300, 1, False)
    store.peer_has_block(std_hash(b"block1"), peer_ids[1], 300, 1, True)
    store.peer_has_block(std_hash(b"block10"), peer_ids[2], 500, 10, False)
    store.peer_has_block(std_hash(b"block1"), peer_ids[2], 300, 1, False)

    assert store.get_heaviest_peak().header_hash == std_hash(b"block10")
    assert store.get_heaviest_peak().height == 10
    assert store.get_heaviest_peak().weight == 500

    assert len(store.get_peak_of_each_peer()) == 2
    store.peer_has_block(std_hash(b"block1"), peer_ids[2], 500, 1, True)
    assert len(store.get_peak_of_each_peer()) == 3
    assert store.get_peak_of_each_peer()[peer_ids[0]].weight == 500
    assert store.get_peak_of_each_peer()[peer_ids[1]].weight == 300
    assert store.get_peak_of_each_peer()[peer_ids[2]].weight == 500

    assert store.get_peers_that_have_peak([std_hash(b"block1")]) == set(peer_ids)
    assert store.get_peers_that_have_peak([std_hash(b"block10")]) == {peer_ids[0], peer_ids[2]}

    store.peer_disconnected(peer_ids[0])
    assert store.get_heaviest_peak().weight == 500
    assert len(store.get_peak_of_each_peer()) == 2
    assert store.get_peers_that_have_peak([std_hash(b"block10")]) == {peer_ids[2]}
    store.peer_disconnected(peer_ids[2])
    assert store.get_heaviest_peak().weight == 300
    store.peer_has_block(std_hash(b"block30"), peer_ids[0], 700, 30, True)
    assert store.get_peak_of_each_peer()[peer_ids[0]].weight == 700
    assert store.get_heaviest_peak().weight == 700


@pytest.mark.anyio
async def test_is_backtrack_syncing_works_when_not_present(seeded_random: random.Random) -> None:
    store = SyncStore()
    node_id = bytes32.random(r=seeded_random)

    assert node_id not in store._backtrack_syncing

    assert not store.is_backtrack_syncing(node_id=node_id)
    assert node_id not in store._backtrack_syncing


@pytest.mark.anyio
async def test_increment_backtrack_syncing_adds(seeded_random: random.Random) -> None:
    store = SyncStore()
    node_id = bytes32.random(r=seeded_random)

    assert node_id not in store._backtrack_syncing

    store.increment_backtrack_syncing(node_id=node_id)
    assert node_id in store._backtrack_syncing


@pytest.mark.anyio
async def test_increment_backtrack_syncing_increments(seeded_random: random.Random) -> None:
    store = SyncStore()
    node_id = bytes32.random(r=seeded_random)

    store.increment_backtrack_syncing(node_id=node_id)
    store.increment_backtrack_syncing(node_id=node_id)
    assert store._backtrack_syncing[node_id] == 2


@pytest.mark.anyio
async def test_decrement_backtrack_syncing_does_nothing_when_not_present(
    seeded_random: random.Random,
) -> None:
    store = SyncStore()
    node_id = bytes32.random(r=seeded_random)

    assert node_id not in store._backtrack_syncing

    store.decrement_backtrack_syncing(node_id=node_id)
    assert node_id not in store._backtrack_syncing


@pytest.mark.anyio
async def test_decrement_backtrack_syncing_decrements(seeded_random: random.Random) -> None:
    store = SyncStore()
    node_id = bytes32.random(r=seeded_random)

    store._backtrack_syncing[node_id] = 2
    store.decrement_backtrack_syncing(node_id=node_id)
    assert store._backtrack_syncing[node_id] == 1


@pytest.mark.anyio
async def test_decrement_backtrack_syncing_removes_at_0(seeded_random: random.Random) -> None:
    store = SyncStore()
    node_id = bytes32.random(r=seeded_random)

    store._backtrack_syncing[node_id] = 1
    store.decrement_backtrack_syncing(node_id=node_id)
    assert node_id not in store._backtrack_syncing


@pytest.mark.anyio
async def test_backtrack_syncing_removes_on_disconnect(seeded_random: random.Random) -> None:
    store = SyncStore()
    node_id = bytes32.random(r=seeded_random)

    assert node_id not in store._backtrack_syncing

    store.increment_backtrack_syncing(node_id=node_id)
    assert node_id in store._backtrack_syncing

    store.peer_disconnected(node_id=node_id)
    assert node_id not in store._backtrack_syncing



from __future__ import annotations

from collections.abc import Awaitable
from typing import Callable, Optional

from chik.types.block_protocol import BlockInfo
from chik.types.blockchain_format.sized_bytes import bytes32
from chik.types.generator_types import BlockGenerator
from chik.util.ints import uint32


async def get_block_generator(
    lookup_block_generators: Callable[[bytes32, set[uint32]], Awaitable[dict[uint32, bytes]]],
    block: BlockInfo,
) -> Optional[BlockGenerator]:
    ref_list = block.transactions_generator_ref_list
    if block.transactions_generator is None:
        assert len(ref_list) == 0
        return None
    if len(ref_list) == 0:
        return BlockGenerator(block.transactions_generator, [])

    generator_refs = set(ref_list)
    generators: dict[uint32, bytes] = await lookup_block_generators(block.prev_header_hash, generator_refs)

    result = [generators[height] for height in block.transactions_generator_ref_list]
    return BlockGenerator(block.transactions_generator, result)



from __future__ import annotations

from typing import Optional

import click

from chik.cmds.cmd_classes import ChikCliContext
from chik.cmds.cmds_util import NODE_TYPES
from chik.cmds.peer_funcs import peer_async


@click.command("peer", help="Show, or modify peering connections", no_args_is_help=True)
@click.option(
    "-p",
    "--rpc-port",
    help=(
        "Set the port where the farmer, wallet, full node or harvester "
        "is hosting the RPC interface. See the rpc_port in config.yaml"
    ),
    type=int,
    default=None,
)
@click.option(
    "-c", "--connections", help="List connections to the specified service", is_flag=True, type=bool, default=False
)
@click.option("-a", "--add-connection", help="Connect specified Chik service to ip:port", type=str, default="")
@click.option(
    "-r", "--remove-connection", help="Remove a Node by the first 8 characters of NodeID", type=str, default=""
)
@click.argument("node_type", type=click.Choice(list(NODE_TYPES.keys())), nargs=1, required=True)
@click.pass_context
def peer_cmd(
    ctx: click.Context,
    rpc_port: Optional[int],
    connections: bool,
    add_connection: str,
    remove_connection: str,
    node_type: str,
) -> None:
    import asyncio

    asyncio.run(
        peer_async(
            node_type,
            rpc_port,
            ChikCliContext.set_default(ctx).root_path,
            connections,
            add_connection,
            remove_connection,
        )
    )



from __future__ import annotations

import os
import sys
from pathlib import Path
from typing import Any, Callable, Optional

import click

from chik.cmds.cmds_util import format_bytes, validate_directory_writable
from chik.util.beta_metrics import metrics_log_interval_max, metrics_log_interval_min
from chik.util.chik_logging import get_beta_logging_config
from chik.util.errors import InvalidPathError


def default_beta_root_path() -> Path:
    return Path(os.path.expanduser(os.getenv("CHIK_BETA_ROOT", "~/chik-beta-test"))).resolve()


def warn_if_beta_enabled(config: dict[str, Any]) -> None:
    if config.get("beta", {}).get("enabled", False):
        print("\nWARNING: beta test mode is enabled. Run `chik beta disable` if this is unintentional.\n")


def prompt_beta_warning() -> bool:
    logging_config = get_beta_logging_config()
    # The `/ 5` is just a rough estimation for `gzip` being used by the log rotation in beta mode. It was like
    # 7-10x compressed in example tests with 2MB files.
    min_space = format_bytes(int(logging_config["log_maxfilesrotation"] * logging_config["log_maxbytesrotation"] / 5))
    message = (
        f"\nWARNING: Enabling the beta test mode increases disk writes and may lead to {min_space} of "
        "extra logfiles getting stored on your disk. This should only be done if you are part of the beta test "
        "program.\n\nDo you really want to enable the beta test mode?"
    )
    return click.confirm(message, default=None)


def prompt_for_beta_path(default_path: Path) -> Path:
    path: Optional[Path] = None
    for _ in range(3):
        user_input = input(
            "\nEnter a directory where the beta test logs can be stored or press enter to use the default "
            f"[{default_path!s}]:"
        )
        test_path = Path(user_input) if user_input else default_path
        if not test_path.is_dir() and click.confirm(
            f"\nDirectory {str(test_path)!r} doesn't exist.\n\nDo you want to create it?",
            default=None,
        ):
            test_path.mkdir(parents=True)

        try:
            validate_directory_writable(test_path)
        except InvalidPathError as e:
            print(str(e))
            continue

        path = test_path
        break

    if path is None:
        sys.exit("Aborted!")
    else:
        return path


def prompt_for_metrics_log_interval(default_interval: int) -> int:
    interval: Optional[int] = None
    for _ in range(3):
        user_input = input(
            "\nEnter a number of seconds as interval in which analytics getting logged, press enter to use the default "
            f"[{default_interval!s}]:"
        )
        test_interval = int(user_input) if user_input else default_interval

        try:
            validate_metrics_log_interval(test_interval)
        except ValueError as e:
            print("\nERROR: " + str(e))
            continue

        interval = test_interval
        break

    if interval is None:
        sys.exit("Aborted!")
    else:
        return interval


def update_beta_config(enabled: bool, path: Path, metrics_log_interval: int, config: dict[str, Any]) -> None:
    if "beta" not in config:
        config["beta"] = {}

    config["beta"].update(
        {
            "enabled": enabled,
            "path": str(path),
            "metrics_log_interval": metrics_log_interval,
        }
    )


def validate_beta_path(beta_root_path: Path) -> None:
    try:
        validate_directory_writable(beta_root_path)
    except InvalidPathError as e:
        sys.exit(str(e))


def validate_metrics_log_interval(interval: int) -> None:
    if interval < metrics_log_interval_min or interval > metrics_log_interval_max:
        raise ValueError(f"Must be in the range of {metrics_log_interval_min}s to {metrics_log_interval_max}s.")


def prepare_plotting_log(path: Path) -> None:
    # TODO: Do stuff we want to do with the logs before submission. Maybe even just fully parse them and
    #  create some final result files and zip them instead of just the logs.
    print(f"  - {path.name}")


def prepare_chik_blockchain_log(path: Path) -> None:
    # TODO: Do stuff we want to do with the logs before submission. Maybe even just fully parse them and
    #  create some final result files and zip them instead of just the logs.
    print(f"  - {path.name}")


def prepare_logs(prepare_path: Path, prepare_callback: Callable[[Path], None]) -> list[Path]:
    result = [path for path in prepare_path.iterdir()] if prepare_path.exists() else []
    if len(result) > 0:
        print(f"\nPreparing {prepare_path.name!r} logs:")
        for log in result:
            if log.name.startswith("."):
                continue
            prepare_callback(log)

    return result
